# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modules that parameterize random processes."""

import abc
from collections.abc import Sequence
import dataclasses
import zlib

from flax import nnx
import jax
import jax.numpy as jnp
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import typing
from neuralgcm.experimental import units
import numpy as np


Quantity = typing.Quantity
_ADVANCE_SALT = zlib.crc32(b'advance')  # arbitrary uint32 value


class RandomnessValue(nnx.Intermediate):
  """Variable type in which randomness values are stored."""


class RandomnessParam(nnx.Variable):
  """Variable type for parameters of the random processes."""


def _advance_prng_key(prng_key: jax.Array, prng_step: int):
  """Get a PRNG Key suitable for advancing randomness from key and step."""
  salt = jnp.uint32(_ADVANCE_SALT) + jnp.uint32(prng_step)
  return jax.random.fold_in(prng_key, salt)


def _get_advance_prng_key(randomness: typing.Randomness) -> jax.Array | None:
  """Extracts advance_prng_key from Randomness."""
  return _advance_prng_key(randomness.prng_key, randomness.prng_step)


class RandomProcessModule(nnx.Module, abc.ABC):
  """Base class for random processes."""

  @abc.abstractmethod
  def unconditional_sample(
      self,
      rng: jax.Array,
  ) -> typing.Randomness:
    """Samples the random field unconditionally."""

  @abc.abstractmethod
  def advance(
      self,
      state: typing.Randomness | None = None,
  ) -> typing.Randomness:
    """Updates the state of a random field."""

  @abc.abstractmethod
  def state_values(
      self,
      coords: cx.Coordinate | None = None,
      state: typing.Randomness | None = None,
  ) -> cx.Field:
    ...

  @property
  def event_shape(self) -> tuple[int, ...]:
    """Shape of the random process event."""
    return tuple()


@dataclasses.dataclass
class UniformUncorrelated(RandomProcessModule):
  """Scalar time-independent uniform random process."""

  coords: cx.Coordinate
  minval: float
  maxval: float

  def unconditional_sample(self, rng):
    key, state_rng = jax.random.split(rng)
    value = jax.random.uniform(
        key,
        minval=self.minval,
        maxval=self.maxval,
        shape=self.coords.shape,
    )
    self.state = RandomnessValue(typing.Randomness(state_rng, 0, value))
    return self.state.value

  def advance(
      self,
      state: typing.Randomness | None = None,
  ) -> typing.Randomness:
    if state is None:
      state = self.state.value
    key = _get_advance_prng_key(state)
    new_value = jax.random.uniform(
        key,
        minval=self.minval,
        maxval=self.maxval,
        shape=self.coords.shape,
    )
    next_state = typing.Randomness(
        state.prng_key, state.prng_step + 1, new_value
    )
    self.state = RandomnessValue(next_state)
    return self.state.value

  def set_state(self, state: typing.Randomness):
    self.state = RandomnessValue(state)

  def get_state(self) -> typing.Randomness:
    return self.state.value

  def state_values(
      self,
      coords: cx.Coordinate | None = None,
      state: typing.Randomness | None = None,
  ) -> cx.Field:
    if state is None:
      state = self.state.value
    if coords is None:
      coords = self.coords
    if coords != self.coords:
      raise ValueError(
          f'Interpolation is not supported yet: {coords=} {self.coords=}'
      )
    return cx.wrap(state.core, coords)


class GaussianRandomFieldCore(nnx.Module):
  """Core functionality of a spatio-temporal gaussian random process.

  This is a core class that is used to define multiple RandomProcessModules, but
  is not a RandomProcessModule itself. The rationale for parameterizing the core
  logic separately is to avoid nesting in RandomProcessModule classes, which
  could interfere with global initialization of the randomness.

  For implementation details see Appendix 8 in [1].
    [1] Stochastic Parametrization and Model Uncertainty, Palmer Et al. 2009.

  With x ∈ EarthSurface, this field U is initialized at t=0 with
    U(0, x) = Σₖ Ψₖ(x) (1 - ϕ²)^(-0.5) σₖ γₖ σₖ ηₖ₀,
  where Ψₖ is the kth spherical harmonic basis function, ϕ² is the one timestep
  correlation, σₖ > 0 is a scaling factor, and ηₖ₀ are iid 1D unit Gaussians.

  With `variance` an init kwarg,
    E[U(0, x)] ≡ 0,
    1 / (4πR²) ∫ Var(U(0, x))dx = variance,
  regardless of coords (and the radius).

  Further states are generated with the recursion
    U(t + δ) = ϕ U(t) + σₖ ηₖₜ
  This ensures that U is stationary.

  In general, the j timestep correlation is
    Cov(U(t, x), U(t + jδ, y)) = ϕ²ʲ Σₖ Ψₖ(x) Ψₖ(y) (γₖ)²
  """

  def __init__(
      self,
      grid: coordinates.LonLatGrid,
      dt: float,
      sim_units: units.SimUnits,
      correlation_time: typing.Numeric | typing.Quantity,
      correlation_length: typing.Numeric | typing.Quantity,
      variance: typing.Numeric,
      correlation_time_type: nnx.Param | RandomnessParam = RandomnessParam,
      correlation_length_type: nnx.Param | RandomnessParam = RandomnessParam,
      variance_type: nnx.Param | RandomnessParam = RandomnessParam,
      clip: float = 6.0,
  ):
    """Constructs a core of a Gaussian Random Field.

    Args:
      grid: lon-lat coordinate system on which the random process is defined.
      dt: time step of the random process.
      sim_units: object defining nondimensionalization and physical constants.
      correlation_time: correlation time of the random process.
      correlation_length: correlation length of the random process.
      variance: variance of the random process.
      correlation_time_type: parameter type for correlation time that allows for
        granular selection of the subsets of model parameters.
      correlation_length_type: parameter type for correlation length that allows
        for granular selection of the subsets of model parameters.
      variance_type: parameter type for variance  that allows for granular
        selection of the subsets of model parameters.
      clip: number of standard deviations at which to clip randomness to ensure
        numerical stability.
    """
    nondimensionalize = lambda x: units.maybe_nondimensionalize(x, sim_units)
    correlation_time = nondimensionalize(correlation_time)
    correlation_length = nondimensionalize(correlation_length)
    variance = nondimensionalize(variance)
    # we make parameters 1d to streamline broadcasting when code is vmapped.
    as_1d_param = lambda x, t: t(jnp.array([x]))
    self.grid = grid
    self.dt = dt
    self.corr_time = as_1d_param(correlation_time, correlation_time_type)
    self.corr_length = as_1d_param(correlation_length, correlation_length_type)
    self._variance = as_1d_param(variance, variance_type)
    self.clip = clip

  @property
  def _surf_area(self) -> jax.Array | float:
    """Surface area of the sphere used by self.grid."""
    return 4 * jnp.pi * self.grid.radius**2

  @property
  def variance(self):
    """An estimate of pointwise (in nodal space) variance of this random field.

    This random field is defined in spectral space, and has no precise
    pointwise variance quantity. However, it does have a precise integrated
    variance, which is used to define the field.

    If we assume the field is stationary (with higher spectral
    precision it is near stationary), then the average of this quantity is a
    good pointwise estimate. So define
      σ² := (1 / (4πR²)) ∫ Var(U(0, x))dx
          = (1 / (4πR²)) integrated_grf_variance

    Therefore the init parameter `variance` can be used to define
      `_integrated_grf_variance := variance * surf_area`
    and then `_integrated_grf_variance` is used to define this field. The result
    is a field with pointwise variance close to the init kwarg `variance`.

    Returns:
      Numeric estimate of pointwise variance.
    """
    return self._variance.value

  @property
  def phi(self) -> jax.Array:
    """Correlation coefficient between two timesteps."""
    return jnp.exp(-self.dt / self.corr_time.value)

  @property
  def relative_corr_len(self):
    """Correlation length of the random process relative to the radius."""
    return self.corr_length.value / self.grid.radius

  def _integrated_grf_variance(self):
    """Integral of the GRF's variance over the earth's surface."""
    return self.variance * self._surf_area

  def _sigma_array(self) -> jax.Array:
    """Array of σₙ from Appendix 8 in [Palmer] http://shortn/_56HCcQwmSS."""
    dinosaur_grid = self.grid.ylm_grid
    # n = [0, 1, ..., N]
    n = dinosaur_grid.modal_axes[1]  # total wavenumbers.
    # Number of longitudinal wavenumbers at each total wavenumber n.
    #  L = 2n + 1, except for the last entry.
    n_longitudian_wavenumbers = dinosaur_grid.mask.sum(axis=0)
    # [Palmer] states correlation_length = sqrt(2κT) / R, therefore
    kt = 0.5 * self.relative_corr_len ** 2
    # sigmas_unnormed[n] is proportional to the standard deviation for each
    # longitudinal wavenumbers at each total wavenumber n.
    sigmas_unnormed = jnp.exp(-0.5 * kt * n * (n + 1))
    # The sum of unnormalized variance for all longitudinal wavenumbers at each
    # total wavenumber.
    sum_unnormed_vars = jnp.sum(n_longitudian_wavenumbers * sigmas_unnormed**2)
    # This is analogous to F₀ from [Palmer].
    # (normalization * sigmas_unnormed)² would sum to 1. The leading factor
    #   self._integrated_grf_variance * (1 - self.phi ** 2)
    # ensures that the AR(1) process has variance self._integrated_grf_variance.
    # We do not include the extra fator of 2 in the denominator. I do not know
    # why [Palmer] has this factor.
    # In sampling, phi appears as 1 - phi**2 = 1 - exp(-2 dt / tau)
    one_minus_phi2 = -jnp.expm1(-2 * self.dt / self.corr_time.value)
    normalization = jnp.sqrt(
        self._integrated_grf_variance() * one_minus_phi2 / sum_unnormed_vars
    )
    # The factor of coords.horizontal.radius appears because our basis vectors
    # have L2 norm = radius.
    return normalization * sigmas_unnormed / dinosaur_grid.radius

  def sample_core(self, rng: typing.PRNGKeyArray) -> jax.Array:
    """Helper method for sampling the core of the gaussian random field."""
    dinosaur_grid = self.grid.ylm_grid
    modal_shape = dinosaur_grid.modal_shape
    sigmas = self._sigma_array()
    weights = jnp.where(
        dinosaur_grid.mask,
        jax.random.truncated_normal(rng, -self.clip, self.clip, modal_shape),
        jnp.zeros(modal_shape),
    )
    one_minus_phi2 = -jnp.expm1(-2 * self.dt / self.corr_time.value)
    return one_minus_phi2 ** (-0.5) * sigmas * weights

  def advance_core(
      self, state_core: jax.Array, state_key: jax.Array, state_step: int
  ) -> jax.Array:
    """Helper method for advancing the core of the gaussian random field."""
    dinosaur_grid = self.grid.ylm_grid
    modal_shape = dinosaur_grid.modal_shape
    rng = _advance_prng_key(state_key, state_step)
    eta = jax.random.truncated_normal(rng, -self.clip, self.clip, modal_shape)
    return state_core * self.phi + self._sigma_array() * jnp.where(
        dinosaur_grid.mask, eta, jnp.zeros(modal_shape)
    )


class GaussianRandomField(RandomProcessModule):
  """Spatially and temporally correlated gaussian process on the sphere."""

  def __init__(
      self,
      grid: coordinates.LonLatGrid,
      dt: float,
      sim_units: units.SimUnits,
      correlation_time: typing.Numeric | typing.Quantity,
      correlation_length: typing.Numeric | typing.Quantity,
      variance: typing.Numeric,
      correlation_time_type: nnx.Param | RandomnessParam = RandomnessParam,
      correlation_length_type: nnx.Param | RandomnessParam = RandomnessParam,
      variance_type: nnx.Param | RandomnessParam = RandomnessParam,
      clip: float = 6.0,
      *,
      rngs: nnx.Rngs,
  ):
    """Initializes a Gaussian Random Field."""
    self.grf = GaussianRandomFieldCore(
        grid=grid,
        dt=dt,
        sim_units=sim_units,
        correlation_time=correlation_time,
        correlation_length=correlation_length,
        variance=variance,
        correlation_time_type=correlation_time_type,
        correlation_length_type=correlation_length_type,
        variance_type=variance_type,
        clip=clip,
    )
    # rngs are used to create an initial state. This ensures that the structure
    # of the module is not changed by sampling or advancing the state.
    prng_key = rngs.params()._base_array
    core_rngs = rngs.params()._base_array
    self.randomness_state = RandomnessValue(
        typing.Randomness(
            prng_key=prng_key,
            prng_step=0,
            core=self.grf.sample_core(core_rngs),
        )
    )

  def unconditional_sample(self, rng: typing.PRNGKeyArray) -> typing.Randomness:
    """Returns a randomly initialized state for the autoregressive process."""
    rng, next_rng = jax.random.split(rng)
    randomness_state = typing.Randomness(
        prng_key=next_rng,
        prng_step=0,
        core=self.grf.sample_core(rng),
    )
    self.randomness_state = RandomnessValue(randomness_state)
    return randomness_state

  def advance(
      self, state: typing.Randomness | None = None
  ) -> typing.Randomness:
    """Updates the CoreRandomState of a random gaussian field."""
    if state is None:
      state = self.randomness_state.value
    next_randomness_state = typing.Randomness(
        core=self.grf.advance_core(state.core, state.prng_key, state.prng_step),
        prng_key=state.prng_key,
        prng_step=state.prng_step + 1,
    )
    self.randomness_state = RandomnessValue(next_randomness_state)
    return next_randomness_state

  def state_values(
      self,
      coords: cx.Coordinate | None = None,
      state: typing.Randomness | None = None,
  ) -> cx.Field:
    if coords is None:
      coords = self.grid
    if not isinstance(coords, coordinates.LonLatGrid):
      raise ValueError('Interpolation of randomness is not supported yet')
    if state is None:
      state = self.randomness_state.value
    dinosaur_grid = self.grf.grid.ylm_grid
    return cx.wrap(dinosaur_grid.to_nodal(state.core), coords)


class BatchGaussianRandomField(RandomProcessModule):
  """Batched version of GaussianRandomField process."""

  def __init__(
      self,
      grid: coordinates.LonLatGrid,
      dt: float,
      sim_units: units.SimUnits,
      correlation_times: typing.Numeric | typing.Quantity,
      correlation_lengths: typing.Numeric | typing.Quantity,
      variances: Sequence[float],
      correlation_time_type: nnx.Param | RandomnessParam = RandomnessParam,
      correlation_length_type: nnx.Param | RandomnessParam = RandomnessParam,
      variance_type: nnx.Param | RandomnessParam = RandomnessParam,
      clip: float = 6.0,
      *,
      rngs: nnx.Rngs,
  ):
    lengths = [
        len(correlation_times),
        len(correlation_lengths),
        len(variances),
    ]
    if len(set(lengths)) != 1:
      raise ValueError(f'Argument lengths differed: {lengths=}')
    (n_fields,) = set(lengths)

    nondimensionalize = lambda x: units.maybe_nondimensionalize(x, sim_units)
    correlation_times = np.array(
        [nondimensionalize(tau) for tau in correlation_times]
    )
    correlation_lengths = np.array(
        [nondimensionalize(length) for length in correlation_lengths]
    )
    variances = np.array(
        [nondimensionalize(variance) for variance in variances]
    )
    make_grf = lambda length, tau, variance: GaussianRandomFieldCore(
        grid=grid,
        dt=dt,
        sim_units=sim_units,
        correlation_time=tau,
        correlation_length=length,
        variance=variance,
        correlation_time_type=correlation_time_type,
        correlation_length_type=correlation_length_type,
        variance_type=variance_type,
        clip=clip,
    )
    self.n_fields = n_fields
    self.batch_grf_core = nnx.vmap(make_grf, axis_size=self.n_fields)(
        correlation_lengths, correlation_times, variances
    )
    prng_key = rngs.params()._base_array
    core_rngs = jax.random.split(rngs.params()._base_array, num=self.n_fields)
    sample_fn = lambda grf, rng: grf.sample_core(rng)
    self.randomness_state = RandomnessValue(
        typing.Randomness(
            prng_key=prng_key,
            prng_step=0,
            core=nnx.vmap(sample_fn)(self.batch_grf_core, core_rngs),
        )
    )

  @property
  def event_shape(self) -> tuple[int, ...]:
    return tuple([self.n_fields])

  def unconditional_sample(self, rng: typing.PRNGKeyArray) -> typing.Randomness:
    all_rngs = jax.random.split(rng, num=(self.n_fields + 1))
    next_rng = all_rngs[-1]
    rngs = all_rngs[:-1]
    sample_fn = lambda grf, rng: grf.sample_core(rng)
    randomness_state = typing.Randomness(
        prng_key=next_rng,
        prng_step=0,
        core=nnx.vmap(sample_fn)(self.batch_grf_core, rngs),
    )
    self.randomness_state = RandomnessValue(randomness_state)
    return randomness_state

  def advance(
      self, state: typing.Randomness | None = None
  ) -> typing.Randomness:
    if state is None:
      state = self.state.value
    advance_fn = lambda grf, core, key, step: grf.advance_core(core, key, step)
    in_axes = (0, 0, 0, None)
    advance_fn = nnx.vmap(advance_fn, in_axes=in_axes)
    rngs = jax.random.split(state.prng_key, num=self.n_fields)
    next_randomness_state = typing.Randomness(
        core=advance_fn(self.batch_grf_core, state.core, rngs, state.prng_step),
        prng_key=state.prng_key,
        prng_step=state.prng_step + 1,
    )
    self.randomness_state = RandomnessValue(next_randomness_state)
    return next_randomness_state

  def state_values(
      self,
      coords: cx.Coordinate | None = None,
      state: typing.Randomness | None = None,
  ) -> cx.Field:
    if coords is None:
      coords = self.grid
    if not isinstance(coords, coordinates.LonLatGrid):
      raise ValueError('Interpolation of randomness is not supported yet')
    if state is None:
      state = self.randomness_state.value
    dinosaur_grid = self.batch_grf_core.grid.ylm_grid
    # TODO(dkochkov): consider an option where we return a field with only
    # trailing axes labeled, rather than assigning names to all dimensions.
    return cx.wrap(dinosaur_grid.to_nodal(state.core), 'grf', coords)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for initializers.py."""

from typing import Tuple
from absl.testing import absltest
from absl.testing import parameterized
import haiku as hk
import jax
from neuralgcm import initializers
import numpy as np


class InitializaersTest(parameterized.TestCase):
  """Tests initializer primitives."""

  @parameterized.parameters(
      dict(shape=(100, 100), scale=0.1),
      dict(shape=(200, 500), scale=10.5),
      dict(shape=(3000, 2000), scale=1.0),
  )
  def test_reducing_variance_initializer_values(
      self,
      shape: Tuple[int, ...],
      scale: float,
  ):
    """Tests ReducingVarianceScaling initialization."""

    @hk.transform
    def get_init_params(shape, dtype):
      init = initializers.ReducingVarianceScaling(scale)
      p = hk.get_parameter('p', shape, dtype, init=init)
      return p

    params = get_init_params.init(jax.random.PRNGKey(42), shape, np.float32)
    params = jax.tree_util.tree_leaves(params)[0]
    actual_scaled_variance = shape[0] * np.std(params.sum(0))**2
    expected_scaled_variance = scale
    np.testing.assert_allclose(
        actual_scaled_variance, expected_scaled_variance, rtol=0.1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modules that parameterize composed time-steppers."""

import abc
import functools
from typing import Any, Callable, Optional, Sequence
from dinosaur import coordinate_systems
from dinosaur import primitive_equations
from dinosaur import time_integration
from dinosaur import typing
import gin
import haiku as hk
from neuralgcm import diagnostics
from neuralgcm import integrators
from neuralgcm import perturbations
from neuralgcm import stochastic

DiagnosticModule = diagnostics.DiagnosticModule
Forcing = typing.Forcing
Pytree = typing.Pytree
ModelState = typing.ModelState
EquationModule = Callable[..., time_integration.ImplicitExplicitODE]
CorrectorModule = typing.CorrectorModule
PerturbationModule = perturbations.PerturbationModule
RandomnessModule = stochastic.RandomnessModule
PyTreeStepFilterModule = typing.PyTreeStepFilterModule
TimeIntegrator = integrators.TimeIntegrator
TransformModule = typing.TransformModule


class BaseStep(abc.ABC):
  """Base class for Step modules."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      diagnostics_module: DiagnosticModule = diagnostics.NoDiagnostics,
      randomness_module: RandomnessModule = stochastic.NoRandomField,
  ):
    self.diagnostics_fn = diagnostics_module(
        coords, dt, physics_specs, aux_features)
    self.randomness_fn = randomness_module(
        coords, dt, physics_specs, aux_features)

  @abc.abstractmethod
  def __call__(
      self,
      state: ModelState,
      forcing: typing.Forcing,
  ) -> ModelState:
    """Computes the state of the system evolved in time by `self.dt`."""

  def finalize_state(
      self,
      x: ModelState,
      forcing: typing.Forcing,
  ) -> ModelState:
    """Finalizes initialization of a model state `x`, encoded from data.

    This method ensures that state has all of the `ModelState` fields
    initialized in a way compatible with this step function. This includes
    populating initial `diagnostics`, `memory` and `randomness` fields.

    Args:
      x: Initial values for the model state typically provided by the encoder.
        forcing: Data covariates from the same time slice as `x`.

    Returns:
      Initialized model state.
    """
    x.randomness = self.randomness_fn.unconditional_sample(
        hk.maybe_next_rng_key()
    )
    x.diagnostics = self.diagnostics_fn(
        x, physics_tendencies=None, forcing=forcing)
    return x


@gin.register
class EquationStep(BaseStep, hk.Module):
  """Step module that advances the state by integrating an equation in time."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      equation_module: EquationModule,
      time_integrator: TimeIntegrator = integrators.imex_rk_sil3,
      filter_modules: Sequence[PyTreeStepFilterModule] = tuple(),
      checkpoint_explicit_terms: bool = True,
      name: Optional[str] = None,
  ):
    hk.Module.__init__(self, name=name)
    BaseStep.__init__(self, coords, dt, physics_specs, aux_features)
    equation = equation_module(coords, dt, physics_specs, aux_features)
    if checkpoint_explicit_terms:
      equation = time_integration.ImplicitExplicitODE.from_functions(
          hk.remat(equation.explicit_terms),
          equation.implicit_terms,
          equation.implicit_inverse)
    step_fn = time_integrator(equation, dt)
    filter_fns = [
        module(coords, dt, physics_specs, aux_features)
        for module in filter_modules]
    self.dt = dt
    self.step_fn = time_integration.step_with_filters(step_fn, filter_fns)

  def __call__(
      self,
      x: ModelState,
      forcing: Optional[typing.Forcing] = None,
  ) -> ModelState:
    """Computes the state of the system evolved in time by `dt`."""
    del forcing
    next_state = time_integration.maybe_fix_sim_time_roundoff(
        self.step_fn(x.state), self.dt)
    return ModelState(next_state)


@gin.register
class RepeatedStep(BaseStep, hk.Module):
  """Step module that consists of repeated substeps of the same form."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      inner_step_module: typing.StepModule,
      num_inner_steps: int = 1,
      name: Optional[str] = None,
  ):
    hk.Module.__init__(self, name=name)
    BaseStep.__init__(self, coords, dt, physics_specs, aux_features)
    inner_dt = dt / num_inner_steps
    self.step_fn = inner_step_module(
        coords, inner_dt, physics_specs, aux_features)
    self.num_inner_steps = num_inner_steps

  def __call__(
      self,
      state: ModelState,
      forcing: typing.Forcing,
  ) -> ModelState:
    """Computes the state of the system evolved in time by `dt`."""
    step_fn = functools.partial(self.step_fn, forcing=forcing)
    step_fn = time_integration.repeated(step_fn, self.num_inner_steps, hk.scan)
    return step_fn(state)


@gin.register
class CustomCoordsStep(BaseStep, hk.Module):
  """Step module that uses gin-configured coordinates instead of coords.

  This class currently supports model states in spectral representation. It
  could be easily extended to nodal-state models by converting to modal space
  prior to spectral interpolation and back after the timestep if performed.
  """

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      step_module: typing.StepModule,
      custom_coords: coordinate_systems.CoordinateSystem = gin.REQUIRED,
      name: Optional[str] = None,
  ):
    hk.Module.__init__(self, name=name)
    BaseStep.__init__(self, coords, dt, physics_specs, aux_features)
    self.step_fn = step_module(
        custom_coords, dt, physics_specs, aux_features)
    self.to_custom_coords_fn = coordinate_systems.get_spectral_interpolate_fn(
        coords, custom_coords)
    self.from_custom_coords_fn = coordinate_systems.get_spectral_interpolate_fn(
        custom_coords, coords)

  def __call__(
      self,
      x: typing.PyTreeState,
      forcing: typing.Forcing,
  ) -> typing.PyTreeState:
    del forcing  # currently not supported.
    x = self.to_custom_coords_fn(x)
    custom_out = self.step_fn(x, None)
    return self.from_custom_coords_fn(custom_out)


@gin.register
class StochasticPhysicsParameterizationStep(BaseStep, hk.Module):
  """Step module that uses stochastic physics tendencies with dycore."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      corrector_module: CorrectorModule,
      physics_parameterization_module: typing.ParameterizationModule,
      num_substeps: int = 1,
      diagnostics_module: DiagnosticModule = diagnostics.NoDiagnostics,
      randomness_module: RandomnessModule = stochastic.ZerosRandomField,
      perturbation_module: PerturbationModule = perturbations.NoPerturbation,
      checkpoint_substep: bool = False,
      name: Optional[str] = None,
  ):
    hk.Module.__init__(self, name=name)
    BaseStep.__init__(
        self, coords, dt, physics_specs, aux_features,
        diagnostics_module=diagnostics_module,
        randomness_module=randomness_module)
    inner_dt = dt / num_substeps
    self.num_substeps = num_substeps
    self.corrector_fn = corrector_module(
        coords, inner_dt, physics_specs, aux_features)
    self.physics_parameterization_fn = physics_parameterization_module(
        coords, inner_dt, physics_specs, aux_features)
    self.perturbation_fn = perturbation_module(
        coords, inner_dt, physics_specs, aux_features)
    self.checkpoint_substep = checkpoint_substep
    self.coords = coords

  def finalize_state(
      self,
      x: ModelState,
      forcing: typing.Forcing,
  ) -> ModelState:
    """Finalizes initialization of a model state `x`, encoded from data.

    This method ensures that state has all of the `ModelState` fields
    initialized in a way compatible with this step function. This includes
    populating initial `diagnostics`, `memory` and `randomness` fields.

    This is called by StochasticModularStepModel.encode, after encoding the data

    Args:
      x: Initial values for the model state typically provided by the encoder.
      forcing: Data covariates from the same time slice as `x`.

    Returns:
      Initialized model state.
    """
    # TODO(dkochkov) Consider adding an option of not overriding randomness.
    x.randomness = self.randomness_fn.unconditional_sample(
        hk.maybe_next_rng_key()
    )
    pp_tendency = self.physics_parameterization_fn(
        x.state, x.memory, x.diagnostics, x.randomness.nodal_value, forcing
    )
    x.diagnostics = self.diagnostics_fn(x, pp_tendency, forcing)
    return x

  def __call__(
      self,
      state: ModelState,
      forcing: typing.Forcing,
  ) -> ModelState:
    """Computes the state of the system evolved in time by `dt`."""

    def step_fn(x):
      x = self.coords.with_dycore_sharding(x)
      # TODO(dkochkov) Consider passing `x` to physics_parameterization.
      pp_tendency = self.physics_parameterization_fn(
          x.state, x.memory, x.diagnostics, x.randomness.nodal_value, forcing
      )

      pp_tendency = self.perturbation_fn(
          pp_tendency,
          state=x.state,
          randomness=x.randomness.nodal_value,
      )

      next_state = self.corrector_fn(x.state, pp_tendency, forcing)
      # TODO(dkochkov) update stochastic modules to take optional state.
      next_randomness = self.randomness_fn.advance(x.randomness)
      next_memory = x.state if x.memory is not None else None
      next_diagnostics = self.diagnostics_fn(x, pp_tendency, forcing)
      x_next = ModelState(
          state=next_state, memory=next_memory, diagnostics=next_diagnostics,
          randomness=next_randomness)
      x_next = self.coords.with_dycore_sharding(x_next)
      return x_next

    if self.checkpoint_substep:
      step_fn = hk.remat(step_fn)
    step_fn = time_integration.repeated(step_fn, self.num_substeps, hk.scan)
    return step_fn(state)


# TODO(dkochkov) Move vertical advection step to transforms.py.


@gin.register
class SemiLagrangianVerticalAdvectionStep(hk.Module):
  """Step module that applies vertical advection for the primitive equations."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.coords = coords
    self.dt = dt

  def __call__(self, state):
    return primitive_equations.semi_lagrangian_vertical_advection_step(
        state, self.coords, self.dt
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Public API for NeuralGCM models."""
from __future__ import annotations

from collections import abc
import datetime
import functools
from typing import Any, Callable

from dinosaur import coordinate_systems
from dinosaur import scales
from dinosaur import time_integration
from dinosaur import typing
from dinosaur import xarray_utils
import jax
from jax import tree_util
import jax.numpy as jnp
from neuralgcm import gin_utils
from neuralgcm import model_builder
from neuralgcm import physics_specifications
import numpy as np
import pandas as pd
import xarray


ArrayLike = float | np.ndarray | jax.Array
Params = dict[str, dict[str, ArrayLike]]
TimedeltaLike = str | np.timedelta64 | pd.Timestamp | datetime.timedelta
Numeric = float | np.ndarray | jax.Array | xarray.DataArray

# TODO(shoyer): make these types more precise
Inputs = dict[str, ArrayLike]
Forcings = dict[str, ArrayLike]
TemporalForcings = dict[str, ArrayLike]
Outputs = dict[str, jax.Array]
BatchedOutputs = dict[str, jax.Array]
State = Any


def _sim_time_from_state(state: State) -> jax.Array:
  """Extract sim_time from model state."""
  # TODO(shoyer): eliminate whichever of these two cases is no longer needed!
  # TODO(shoyer): consider renaming `sim_time` to `time`?
  if isinstance(state, typing.ModelState):
    sim_time = getattr(state.state, 'sim_time', None)
  else:
    sim_time = getattr(state, 'sim_time', None)
  return sim_time


def _calculate_sub_steps(
    timestep: np.timedelta64, duration: TimedeltaLike
) -> int:
  """Calculate the number of time-steps required to simulate a time interval."""
  duration = pd.Timedelta(duration)
  time_step_ratio = duration / timestep
  if abs(time_step_ratio - round(time_step_ratio)) > 1e-6:
    raise ValueError(
        f'non-integral time-step ratio: {duration=} is not a multiple of '
        f'the internal model timestep {timestep}'
    )
  return round(time_step_ratio)


def _prepend_dummy_time_axis(state: typing.Pytree) -> typing.Pytree:
  return tree_util.tree_map(lambda x: jnp.expand_dims(x, axis=0), state)


def _static_gin_config(method):
  """Decorator to add static gin config to a method."""

  @functools.wraps(method)
  def _method(self, *args, **kwargs):
    with gin_utils.specific_config(self.gin_config):
      return method(self, *args, **kwargs)

  return _method


def _check_variables(
    dataset: xarray.Dataset,
    desired_level_variables: abc.Sequence[str] = (),
    desired_surface_variables: abc.Sequence[str] = (),
):
  """Checks that a dataset has the desired variables."""
  T, Z, X, Y = ('time', 'level', 'longitude', 'latitude')  # pylint: disable=invalid-name

  for k in desired_level_variables:
    if k not in dataset.data_vars:
      raise ValueError(f'expected variable {k} not found')
    dims = dataset[k].dims
    if not (set(dims) == {Z, X, Y} or set(dims) == {T, Z, X, Y}):
      raise ValueError(
          f'expected variable {k} to have dims {(Z, X, Y)} or {(T, Z, X, Y)},'
          f' but got {dims}'
      )

  for k in desired_surface_variables:
    if k not in dataset.data_vars:
      raise ValueError(f'expected variable {k} not found')
    dims = dataset[k].dims
    if not (set(dims) == {X, Y} or set(dims) == {T, X, Y}):
      raise ValueError(
          f'expected variable {k} to have dims {(X, Y)} or {(T, X, Y)},'
          f' but got {dims}'
      )


def _check_coords(
    actual_coords: coordinate_systems.CoordinateSystem,
    desired_coords: coordinate_systems.CoordinateSystem,
) -> None:
  """Checks that a dataset has the desired coordinates."""
  if not np.allclose(
      actual := actual_coords.horizontal.longitudes,
      desired := desired_coords.horizontal.longitudes,
      atol=1e-3,
  ):
    raise ValueError(f'longitude coordinate mismatch: {actual=}, {desired=}')

  if not np.allclose(
      actual := actual_coords.horizontal.latitudes,
      desired := desired_coords.horizontal.latitudes,
      atol=1e-3,
  ):
    raise ValueError(f'latitude coordinate mismatch: {actual=}, {desired=}')

  if actual_coords.vertical is not None and not np.allclose(
      actual := actual_coords.vertical.centers,
      desired := desired_coords.vertical.centers,
      atol=1e-3,
  ):
    raise ValueError(
        f'pressure level coordinate mismatch: {actual=}, {desired=}'
    )


def _rename_if_found(
    dataset: xarray.Dataset, names: dict[str, str]
) -> xarray.Dataset:
  return dataset.rename({k: v for k, v in names.items() if k in dataset})


_ABBREVIATED_NAMES = {
    'u_component_of_wind': 'u',
    'v_component_of_wind': 'v',
    'geopotential': 'z',
    'temperature': 't',
    'longitude': 'lon',
    'latitude': 'lat',
}
_FULL_NAMES = {v: k for k, v in _ABBREVIATED_NAMES.items()}


def _expand_tracers(inputs: dict) -> dict:
  inputs = inputs.copy()
  inputs.update(inputs.pop('tracers'))
  assert not inputs['diagnostics']
  del inputs['diagnostics']
  return inputs


@tree_util.register_pytree_node_class
class PressureLevelModel:
  """Inference-only API for models that predict dense data on pressure levels.

  These models are trained on ECMWF ERA5 data on pressure-levels as stored in
  the Copernicus Data Store.

  This class encapsulates the details of defining models (e.g., with Haiku) and
  hence should remain stable even for future NeuralGCM models.
  """

  def __init__(
      self,
      structure: model_builder.WhirlModel,
      params: Params,
      gin_config: str,
  ):
    self._structure = structure
    self._params = params
    self.gin_config = gin_config

    self._tracer_variables = [
        'specific_humidity',
    ]
    self._input_variables = [
        'geopotential',
        'specific_humidity',
        'temperature',
        'u_component_of_wind',
        'v_component_of_wind',
    ]
    # Some old model versions do not use cloud variables.
    # TODO(shoyer): remove this once all integration tests are updated.
    cloud_variables = [
        'specific_cloud_ice_water_content',
        'specific_cloud_liquid_water_content',
    ]
    for variable in cloud_variables:
      if variable in self.gin_config:
        self._tracer_variables.append(variable)
        self._input_variables.append(variable)

    self._forcing_variables = [
        'sea_ice_cover',
        'sea_surface_temperature',
    ]

  def __repr__(self):
    return (
        f'{self.__class__.__name__}(structure={self._structure},'
        f' params={self._params})'
    )

  @property
  def params(self) -> Params:
    return self._params

  def tree_flatten(self):
    leaves, params_def = tree_util.tree_flatten(self.params)
    return (leaves, (params_def, self._structure, self.gin_config))

  @classmethod
  def tree_unflatten(cls, aux_data, leaves):
    params_def, structure, gin_config = aux_data
    params = tree_util.tree_unflatten(params_def, leaves)
    return cls(structure, params, gin_config)

  @property
  def input_variables(self) -> list[str]:
    """List of variable names required in `inputs` by this model."""
    return list(self._input_variables)

  @property
  def forcing_variables(self) -> list[str]:
    """List of variable names required in `forcings` by this model."""
    return list(self._forcing_variables)

  @property
  def timestep(self) -> np.timedelta64:
    """Spacing between internal model timesteps."""
    to_timedelta = (
        self._structure.specs.physics_specs.dimensionalize_timedelta64
    )
    return to_timedelta(self._structure.specs.dt)

  @property
  def data_coords(self) -> coordinate_systems.CoordinateSystem:
    """Coordinate system for input and output data."""
    return self._structure.data_coords

  @property
  def model_coords(self) -> coordinate_systems.CoordinateSystem:
    """Coordinate system for internal model state."""
    return self._structure.coords

  def _check_coords(self, dataset: xarray.Dataset):
    dataset_coords = model_builder.coordinate_system_from_dataset(dataset)
    _check_coords(dataset_coords, self.data_coords)

  def _dataset_with_sim_time(self, dataset: xarray.Dataset) -> xarray.Dataset:
    ref_datetime = self._structure.specs.aux_features['reference_datetime']
    return xarray_utils.ds_with_sim_time(
        dataset,
        self._structure.specs.physics_specs,
        reference_datetime=ref_datetime,
    )

  def _to_abbreviated_names_and_tracers(self, inputs: dict) -> dict:
    inputs = {_ABBREVIATED_NAMES.get(k, k): v for k, v in inputs.items()}
    inputs['tracers'] = {
        k: inputs.pop(k) for k in self._tracer_variables if k in inputs
    }
    inputs['diagnostics'] = {}
    return inputs

  def _from_abbreviated_names_and_tracers(self, outputs: dict) -> dict:
    outputs = {_FULL_NAMES.get(k, k): v for k, v in outputs.items()}
    outputs |= outputs.pop('tracers')
    outputs |= outputs.pop('diagnostics')
    return outputs

  def to_nondim_units(self, value: Numeric, units: str) -> Numeric:
    """Scale a value to the model's internal non-dimensional units."""
    scale_ = self._structure.specs.physics_specs.scale
    units_ = scales.parse_units(units)
    return scale_.nondimensionalize(value * units_)

  def from_nondim_units(self, value: Numeric, units: str) -> Numeric:
    """Scale a value from the model's internal non-dimensional units."""
    scale_ = self._structure.specs.physics_specs.scale
    units_ = scales.parse_units(units)
    return scale_.dimensionalize(value, units_).magnitude

  def datetime64_to_sim_time(self, datetime64: np.ndarray) -> np.ndarray:
    """Converts a datetime64 array to sim_time."""
    ref_datetime = self._structure.specs.aux_features['reference_datetime']
    return xarray_utils.datetime64_to_nondim_time(
        datetime64,
        self._structure.specs.physics_specs,
        reference_datetime=ref_datetime,
    )

  def sim_time_to_datetime64(self, sim_time: np.ndarray) -> np.ndarray:
    """Converts a sim_time array to datetime64."""
    ref_datetime = self._structure.specs.aux_features['reference_datetime']
    return xarray_utils.nondim_time_to_datetime64(
        sim_time,
        self._structure.specs.physics_specs,
        reference_datetime=ref_datetime,
    )

  def _data_from_xarray(
      self, dataset: xarray.Dataset, variables: list[str]
  ) -> dict[str, np.ndarray]:
    self._check_coords(dataset)
    dataset = dataset[variables]
    dataset = self._dataset_with_sim_time(dataset)
    dataset = _rename_if_found(dataset, {'longitude': 'lon', 'latitude': 'lat'})
    return xarray_utils.xarray_to_data_dict(dataset)

  def inputs_from_xarray(
      self, dataset: xarray.Dataset
  ) -> dict[str, np.ndarray]:
    """Extract inputs from an xarray.Dataset."""
    _check_variables(dataset, desired_level_variables=self._input_variables)
    return self._data_from_xarray(dataset, self._input_variables)

  def forcings_from_xarray(
      self, dataset: xarray.Dataset
  ) -> dict[str, np.ndarray]:
    """Extract forcings from an xarray.Dataset."""
    _check_variables(dataset, desired_surface_variables=self._forcing_variables)
    return self._data_from_xarray(dataset, self._forcing_variables)

  def data_from_xarray(
      self, dataset: xarray.Dataset
  ) -> tuple[dict[str, np.ndarray], dict[str, np.ndarray]]:
    """Extracts data and forcings from an xarray.Dataset."""
    inputs = self.inputs_from_xarray(dataset)
    forcings = self.forcings_from_xarray(dataset)
    return (inputs, forcings)

  def data_to_xarray(
      self,
      data: dict[str, ArrayLike],
      times: np.ndarray | None,
      decoded: bool = True,
  ) -> xarray.Dataset:
    """Converts decoded model predictions to xarray.Dataset format.

    Args:
      data: dict of arrays with shapes matching input/outputs or encoded model
        state for this model, i.e., with shape
        `([time,] level, longitude, latitude)`,
        where `[time,]` indicates an optional leading time dimension.
      times: either `None` indicating no leading time dimension on any
        variables, or a coordinate array of times with shape `(time,)`.
      decoded: if `True`, use `self.data_coords` to determine the output
        coordinates; otherwise use `self.model_coords`.

    Returns:
      An xarray.Dataset with appropriate coordinates and dimensions.
    """
    coords = self.data_coords if decoded else self.model_coords
    dataset = xarray_utils.data_to_xarray(data, coords=coords, times=times)
    dataset = _rename_if_found(dataset, {'lon': 'longitude', 'lat': 'latitude'})
    return dataset

  def _squeeze_level_from_forcings(self, forcings: Forcings) -> Forcings:
    # Due to a bug in xarray_to_dynamic_covariate_data, we were accidentally
    # not inserting a level dimension in forcings.
    forcings = dict(forcings)
    for k in self._forcing_variables:
      if k in forcings:
        assert isinstance(forcings[k], (np.ndarray, jax.Array))
        forcings[k] = forcings[k].squeeze(axis=-3)
    return forcings

  @jax.jit
  @_static_gin_config
  def encode(
      self,
      inputs: Inputs,
      forcings: Forcings,
      rng_key: typing.PRNGKeyArray | None = None,
  ) -> State:
    """Encode from pressure-level inputs & forcings to model state.

    Args:
      inputs: input data on pressure-levels, as a dict where each entry is an
        array with shape `[level, longitude, latitude]` matching `data_coords`.
      forcings: forcing data on pressure-levels, as a dict where each entry is
        an array with shape `[level, longitude, latitude]` matching
        `data_coords`. Single level data (e.g., sea surface temperature) should
        have a `level` dimension of size 1.
      rng_key: optional JAX RNG key to use for encoding the state. Required if
        using stochastic models, otherwise ignored.

    Returns:
      Dynamical core state on sigma levels, where all arrays have dimensions
      `[level, zonal_wavenumber, total_wavenumber]` matching `model_coords`.
    """
    sim_time = inputs['sim_time']
    inputs = self._to_abbreviated_names_and_tracers(inputs)
    inputs = _prepend_dummy_time_axis(inputs)
    forcings = self._squeeze_level_from_forcings(forcings)
    forcings = _prepend_dummy_time_axis(forcings)
    f = self._structure.forcing_fn(self.params, None, forcings, sim_time)
    return self._structure.encode_fn(self.params, rng_key, inputs, f)

  @jax.jit
  @_static_gin_config
  def advance(self, state: State, forcings: Forcings) -> State:
    """Advance model state one timestep forward.

    Args:
      state: dynamical core state on sigma levels, where all arrays have
        dimensions `[level, zonal_wavenumber, total_wavenumber]` matching
        `model_coords`
      forcings: forcing data on pressure-levels, as a dict where each entry is
        an array with shape `[level, longitude, latitude]` matching
        `data_coords`. Single level data (e.g., sea surface temperature) should
        have a `level` dimension of size 1.

    Returns:
      State advanced one time-step forward.
    """
    sim_time = _sim_time_from_state(state)
    forcings = self._squeeze_level_from_forcings(forcings)
    forcings = _prepend_dummy_time_axis(forcings)
    f = self._structure.forcing_fn(self.params, None, forcings, sim_time)
    state = self._structure.advance_fn(self.params, None, state, f)
    return state

  @jax.jit
  @_static_gin_config
  def decode(self, state: State, forcings: Forcings) -> Outputs:
    """Decode from model state to pressure-level outputs.

    Args:
      state: dynamical core state on sigma levels, where all arrays have
        dimensions `[level, zonal_wavenumber, total_wavenumber]` matching
        `model_coords`.
      forcings: forcing data on pressure-levels, as a dict where each entry is
        an array with shape `[level, longitude, latitude]` matching
        `data_coords`. Single level data (e.g., sea surface temperature) should
        have a `level` dimension of size 1.

    Returns:
      Outputs on pressure-levels, as a dict where each entry is an array with
      shape `[level, longitude, latitude]` matching `data_coords`.
    """
    sim_time = _sim_time_from_state(state)
    forcings = self._squeeze_level_from_forcings(forcings)
    forcings = _prepend_dummy_time_axis(forcings)
    f = self._structure.forcing_fn(self.params, None, forcings, sim_time)
    outputs = self._structure.decode_fn(self.params, None, state, f)
    outputs = self._from_abbreviated_names_and_tracers(outputs)
    return outputs

  @functools.partial(
      jax.jit,
      static_argnames=[
          'steps',
          'timedelta',
          'start_with_input',
          'post_process_fn',
      ],
  )
  @_static_gin_config
  def unroll(
      self,
      state: State,
      forcings: TemporalForcings,
      *,
      steps: int,
      timedelta: TimedeltaLike | None = None,
      start_with_input: bool = False,
      post_process_fn: Callable[[State], Any] | None = None,
  ) -> tuple[State, BatchedOutputs]:
    """Unroll predictions over many time-steps.

    Usage:

      advanced_state, outputs = model.unroll(state, forcings, steps=N)

    where ``advanced_state`` is the advanced model state after ``N`` steps and
    ``outputs`` is a trajectory of decoded states on pressure-levels with a
    leading dimension of size ``N``.

    Args:
      state: initial model state.
      forcings: forcing data over the time-period spanned by the desired output
        trajectory. Should include a leading time-axis, but times can be at any
        desired granularity (e.g., it should be fine to supply daily forcing
        data, even if producing hourly outputs). The nearest forcing in time
        will be used for each internal ``advance()`` and ``decode()`` call.
      steps: number of time-steps to take.
      timedelta: size of each time-step to take, which must be a multiple of the
        internal model timestep. By default uses the internal model timestep.
      start_with_input: if ``True``, outputs are at times ``[0, ...,
        (steps - 1) * timestep]`` relative to the initial time; if ``False``,
        outputs are at times ``[timestep, ..., steps * timestep]``.
      post_process_fn: optional function to apply to each advanced state and
        current forcings to create outputs like
        ``post_process_fn(state, forcings)``, where ``forcings`` does not
        include a time axis. By default, uses ``model.decode``.

    Returns:
      A tuple of the advanced state at time ``steps * timestamp``, and outputs
      with a leading ``time`` axis at the time-steps specified by ``steps``,
      ``timedelta`` and ``start_with_input``.
    """
    if timedelta is None:
      timedelta = self.timestep

    def get_nearest_forcings(sim_time):
      times = forcings['sim_time']
      assert isinstance(times, jax.Array)
      approx_index = jnp.interp(sim_time, times, jnp.arange(times.size))
      index = jnp.round(approx_index).astype(jnp.int32)
      return jax.tree.map(lambda x: x[index, ...], forcings)

    def with_nearest_forcings(func):
      def wrapped(state):
        sim_time = _sim_time_from_state(state)
        forcings = get_nearest_forcings(sim_time)
        return func(state, forcings)
      return wrapped

    if post_process_fn is None:
      post_process_fn = self.decode

    inner_steps = _calculate_sub_steps(self.timestep, timedelta)
    trajectory_func = time_integration.trajectory_from_step(
        with_nearest_forcings(self.advance),
        outer_steps=steps,
        inner_steps=inner_steps,
        start_with_input=start_with_input,
        post_process_fn=with_nearest_forcings(post_process_fn),
    )
    state, outputs = trajectory_func(state)
    return state, outputs

  @classmethod
  def from_checkpoint(cls, checkpoint: Any) -> PressureLevelModel:
    """Creates a PressureLevelModel from a checkpoint.

    Args:
      checkpoint: dictionary with keys "model_config_str", "aux_ds_dict" and
        "params" that specifies model gin configuration, supplemental xarray
        dataset with model-specific static features, and model parameters.

    Returns:
      Instance of a `PressureLevelModel` with weights and configuration
      specified by the checkpoint.
    """
    with gin_utils.specific_config(checkpoint['model_config_str']):
      physics_specs = physics_specifications.get_physics_specs()
      aux_ds = xarray.Dataset.from_dict(checkpoint['aux_ds_dict'])
      data_coords = model_builder.coordinate_system_from_dataset(aux_ds)
      model_specs = model_builder.get_model_specs(
          data_coords, physics_specs, {xarray_utils.XARRAY_DS_KEY: aux_ds}
      )
      whirl_model = model_builder.WhirlModel(
          coords=model_specs.coords,
          dt=model_specs.dt,
          physics_specs=model_specs.physics_specs,
          aux_features=model_specs.aux_features,
          input_coords=data_coords,
          output_coords=data_coords,
      )
      return cls(
          whirl_model, checkpoint['params'], checkpoint['model_config_str']
      )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests time integrator modules and supporting utilities."""

import dataclasses

from absl.testing import absltest
from absl.testing import parameterized
from flax import nnx
from neuralgcm.experimental import time_integrators


@dataclasses.dataclass
class MockStep(nnx.Module):
  """Mock step module for testing."""

  delta: nnx.Param = dataclasses.field(default_factory=lambda: nnx.Param(0.5))
  cumulative: nnx.Intermediate = dataclasses.field(
      default_factory=lambda: nnx.Intermediate(0.0)
  )
  last_output: nnx.Intermediate = dataclasses.field(
      default_factory=lambda: nnx.Intermediate(0.0)
  )

  def half_increment(self, x):
    output = x + self.delta.value
    self.last_output = nnx.Intermediate(output)
    return output

  def __call__(self, x):
    result = self.half_increment(self.half_increment(x))
    self.cumulative = nnx.Intermediate(self.cumulative.value + result)
    return result


# TODO(dkochkov) Consider using namedtuple for state to improve readability.
class TimeIntegrationUtilsTest(parameterized.TestCase):
  """Tests for time integration utility function."""

  @parameterized.named_parameters(
      dict(
          testcase_name='nnx_module',
          step_fn=lambda mc: (mc[0], mc[0](mc[1])),
          get_input=lambda model, c: (model, c),
          get_expected_output=lambda mc: mc[1],
          inner_steps=3,
          expected_result=3.0,
          expected_state_vars={'cumulative': 6.0, 'last_output': 3.0},
      ),
      dict(
          testcase_name='nnx_module_method',
          step_fn=lambda mc: (mc[0], mc[0].half_increment(mc[1])),
          get_input=lambda model, c: (model, c),
          get_expected_output=lambda mc: mc[1],
          inner_steps=3,
          expected_result=1.5,
          expected_state_vars={'cumulative': 0.0, 'last_output': 1.5},
      ),
      dict(
          testcase_name='pure_function',
          step_fn=lambda c: c + 1,
          get_input=lambda model, c: c,
          get_expected_output=lambda c: c,
          inner_steps=3,
          expected_result=3.0,
          expected_state_vars=None,
      ),
  )
  def test_repeated(
      self,
      step_fn,
      get_input,
      get_expected_output,
      inner_steps: int,
      expected_result: float,
      expected_state_vars: dict[str, float] | None,
  ):
    """Tests output_shape, number and initialization of params in MlpUniform."""
    model = MockStep()
    repeated_step_fn = time_integrators.repeated(step_fn, inner_steps)
    inputs = get_input(model, 0.0)
    outputs = repeated_step_fn(inputs)

    with self.subTest('expected_result'):
      actual = get_expected_output(outputs)
      self.assertAlmostEqual(actual, expected_result)

    if expected_state_vars is not None:
      for k, expected in expected_state_vars.items():
        with self.subTest(f'expected_state_var_{k}'):
          actual = nnx.state(model, nnx.Intermediate)[k].value
          self.assertAlmostEqual(actual, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=line-too-long
# pyformat: disable
"""Xarray based readers for feeding time-series into tf.data."""
# pyformat: enable
from __future__ import annotations
from collections import abc
import concurrent.futures
import dataclasses
import logging
import math
import random
from typing import Callable, Optional, TypeVar

import numpy as np
import tensorflow as tf
import xarray


# pylint: disable=logging-fstring-interpolation


def _xarray_bytes_per_element(
    source: xarray.Dataset, exclude_dims: set[str]
) -> int:
  bytes_per_element = 0
  for variable in source.values():
    items_per_element = math.prod(
        size for dim, size in variable.sizes.items() if dim not in exclude_dims
    )
    bytes_per_element += variable.dtype.itemsize * items_per_element
  return bytes_per_element


def _calculate_block_size(
    source: xarray.Dataset,
    block_dims: list[str],
    bytes_per_request: float,
    min_elements_per_request: int = 1,
) -> int:
  """Calculate the size of blocks to read simultaneously from disk."""
  bytes_per_element = _xarray_bytes_per_element(source, set(block_dims))
  elements_per_request = round(bytes_per_request / bytes_per_element)
  max_elements = math.prod(source.sizes[dim] for dim in block_dims)
  elements_per_request = min(
      max(elements_per_request, min_elements_per_request), max_elements
  )
  return elements_per_request


def _iterate_windowed_block_slices(
    sample_size: int,
    total_size: int,
    block_size: int,
    stride_between_samples: int = 1,
    output_window_stride: int = 1,
    first_sample_offset: int = 0,
) -> abc.Iterator[slice]:
  """Yields slices for every block needed to generate windowed samples.

  Args:
    sample_size: size of each sample.
    total_size: total size of the dimension being sampled along.
    block_size: desired size of blocks to read from disk.
    stride_between_samples: shift between starts of sampled windows.
    output_window_stride: shift between samples within a window.
    first_sample_offset: offset of the first sample.

  Yields:
    Slice objects with integer bounds for each block.
  """
  assert stride_between_samples >= 1
  assert output_window_stride >= 1
  assert first_sample_offset >= 0

  sample_input_size = (
      range(0, sample_size * output_window_stride, output_window_stride)[-1] + 1
  )
  assert 0 < sample_input_size <= block_size <= total_size

  sample_stop = 0  # unused

  # first block
  block_start = first_sample_offset
  block_stop = first_sample_offset + block_size

  # iterate through all slices, in order
  for start in range(
      first_sample_offset,
      total_size - sample_input_size + 1,
      stride_between_samples,
  ):
    prev_sample_stop = sample_stop
    sample_stop = start + sample_input_size

    if sample_stop > block_stop:
      # yield previous block
      assert prev_sample_stop > 0
      yield slice(block_start, prev_sample_stop)

      # begin new block
      block_start = start
      block_stop = start + block_size

  if sample_stop > block_start:
    # yield the final block
    yield slice(block_start, sample_stop)


def _drop_static_vars(dataset: xarray.Dataset) -> xarray.Dataset:
  """Drop fields that are static and do not vary with time."""
  vars_to_drop = [k for k, v in dataset.items() if 'time' not in v.dims[0]]  # pytype: disable=unsupported-operands
  return dataset.drop_vars(vars_to_drop)


NestedTensors = TypeVar('NestedTensors', tf.Tensor, dict[str, tf.Tensor])


@tf.function(jit_compile=True, autograph=False)
def rolling_window_tensors(
    inputs: NestedTensors, /, size: int, shift: int = 1, stride: int = 1
) -> NestedTensors:
  """Calculate a tensor of rolling windows.

  Example usage:

    >>> rolling_window_tensors(tf.range(10), size=6, shift=2)
    <tf.Tensor: shape=(3, 6), dtype=int32, numpy=
    array([[0, 1, 2, 3, 4, 5],
          [2, 3, 4, 5, 6, 7],
          [4, 5, 6, 7, 8, 9]], dtype=int32)>

    >>> rolling_window_tensors(tf.range(10), size=4, stride=2)
    <tf.Tensor: shape=(4, 4), dtype=int32, numpy=
    array([[0, 2, 4, 6],
          [1, 3, 5, 7],
          [2, 4, 6, 8],
          [3, 5, 7, 9]], dtype=int32)>

  Args:
    inputs: nested data structure with tf.Tensor values of shape [T, ...].
    size: size of the time dimension in rolling window samples.
    shift: shift between subsequent window samples along time.
    stride: shift within a window along time.

  Returns:
    Nested tensors of shape [S, W, ...] sampled from inputs, where S is the
    number of samples and W is the window size.
  """

  def calculate_windows(tensor):
    shifts = tf.range(0, tf.shape(tensor)[0] - stride * (size - 1), shift)
    indices = tf.range(0, size * stride, stride)
    samples = tf.vectorized_map(
        lambda shift: tf.gather(tensor, shift + indices), shifts
    )
    samples = tf.ensure_shape(samples, [None, size] + tensor.shape[1:])
    return samples

  return tf.nest.map_structure(calculate_windows, inputs)


class Sampler:
  """Base class for sampling from blocks."""

  def list_block_slices(self, block_size: int, total_size: int) -> list[slice]:
    """Returns a list of slices bounding blocks to sample from."""
    raise NotImplementedError

  def sample_block(self, data: NestedTensors) -> NestedTensors:
    """Returns sample tensors from block tensors."""
    raise NotImplementedError

  @property
  def example_size(self) -> int:
    """Size of each example."""
    raise NotImplementedError

  def examples_per_block(self, block_size: int) -> int:
    """Number of examples per block."""
    raise NotImplementedError


@dataclasses.dataclass
class Splitter(Sampler):
  """Split samples along the first axis."""


  def list_block_slices(self, block_size: int, total_size: int) -> list[slice]:
    return [
        slice(start, min(start + block_size, total_size))
        for start in range(0, total_size, block_size)
    ]

  def sample_block(self, data: NestedTensors) -> NestedTensors:
    # Insert a dummy dimension for time-series length, which is always one.
    return tf.nest.map_structure(lambda x: x[:, tf.newaxis, ...], data)

  @property
  def example_size(self) -> int:
    return 1

  def examples_per_block(self, block_size: int) -> int:
    return block_size


@dataclasses.dataclass
class Windower(Sampler):
  """Sample rolling windows along the first axis.

  Attributes:
    window_size: size of output windows.
    stride_between_windows: offset between starting sequential windows.
    output_window_stride: separation between between observations within a
      window.
    first_window_offset: offset of starting the first window.
  """

  window_size: int
  stride_between_windows: int
  output_window_stride: int = 1
  first_window_offset: int = 0

  def list_block_slices(self, block_size: int, total_size: int) -> list[slice]:
    return list(
        _iterate_windowed_block_slices(
            sample_size=self.window_size,
            block_size=block_size,
            total_size=total_size,
            stride_between_samples=self.stride_between_windows,
            output_window_stride=self.output_window_stride,
            first_sample_offset=self.first_window_offset,
        )
    )

  def sample_block(self, data: NestedTensors) -> NestedTensors:
    # NOTE(shoyer): It is tempting to try to use tf.data.Dataset.window instead
    # for sampling windows, but that method does something different: it
    # calculates windows over Dataset elements, rather than calculating windows
    # within each Dataset element.
    return rolling_window_tensors(
        data,
        size=self.window_size,
        shift=self.stride_between_windows,
        stride=self.output_window_stride,
    )

  @property
  def example_size(self) -> int:
    return self.window_size

  def examples_per_block(self, block_size: int) -> int:
    stop = block_size - self.output_window_stride * (self.window_size - 1)
    return len(range(0, stop, self.stride_between_windows))


@dataclasses.dataclass
class WindowerAtOffsets(Sampler):
  """Sample rolling windows along the first axis at specified offsets."""

  window_size: int
  window_offsets: list[int]
  output_window_stride: int = 1

  def list_block_slices(self, block_size: int, total_size: int) -> list[slice]:
    stride = self.output_window_stride
    # This suffices for now because generally we cache evaluation data.
    sample_input_size = range(0, self.window_size * stride, stride)[-1] + 1
    assert 0 < sample_input_size <= block_size <= total_size
    slices = []
    for start in self.window_offsets:
      stop = start + sample_input_size
      if stop > total_size:
        raise ValueError(
            f'offset at {start} needs data through {stop=}, which is beyond'
            f' {total_size=}'
        )
      slices.append(slice(start, stop))
    return slices

  def sample_block(self, data: NestedTensors) -> NestedTensors:
    def strided_sample(tensor):
      # Insert a dummy batch/sample dimension.
      return tf.ensure_shape(
          tensor[tf.newaxis, :: self.output_window_stride],
          [1, self.window_size] + tensor.shape[1:],
      )

    return tf.nest.map_structure(strided_sample, data)

  @property
  def example_size(self) -> int:
    return self.window_size

  def examples_per_block(self, block_size: int) -> int:
    del block_size  # unused
    return 1


class Selector:
  """Base class for block selection."""

  def select(self, blocks: list[slice]) -> list[slice]:
    """Select a subset of blocks for sampling."""
    raise NotImplementedError


class CompleteSelector(Selector):

  def select(self, blocks: list[slice]) -> list[slice]:
    return blocks


@dataclasses.dataclass
class ShardSelector(Selector):
  shard_index: int
  shard_count: int

  def select(self, blocks: list[slice]) -> list[slice]:
    return [
        block
        for i, block in enumerate(blocks)
        if i % self.shard_count == self.shard_index
    ]


@dataclasses.dataclass
class ShuffleSelector(Selector):
  seed: int = 0
  reshuffle_each_iteration: bool = True

  def select(self, blocks: list[slice]) -> list[slice]:
    rng = random.Random(self.seed)
    if self.reshuffle_each_iteration:
      self.seed = rng.randrange(2**63)
    return rng.sample(blocks, k=len(blocks))


@dataclasses.dataclass
class ComposedSelector(Selector):
  components: list[Selector]

  def select(self, blocks: list[slice]) -> list[slice]:
    for component in self.components:
      blocks = component.select(blocks)
    return blocks


@dataclasses.dataclass
class CustomSelector(Selector):
  select: Callable[[list[slice]], list[slice]]


def _thread_pool_loader(max_workers: int = 100):
  """Dataset loader using a large thread pool for concurrency."""
  # We use a separate thread for reading each data variable in each block.
  executor = concurrent.futures.ThreadPoolExecutor(max_workers)

  def load(dataset: xarray.Dataset) -> xarray.Dataset:
    arrays = executor.map(lambda var: var.values, dataset.values())
    return dataset.copy(data={k: v for k, v in zip(dataset, arrays)})

  return load


class _Reader:
  """Class for reading an xarray.Dataset."""

  def __init__(
      self,
      source: xarray.Dataset,
      sampler: Sampler,
      block_selector: Selector = CompleteSelector(),
      *,
      sample_dim: str = 'time',
      block_size_in_bytes: float = 1e8,
      parallel_block_reads: int = tf.data.AUTOTUNE,
      parallel_samples: int = tf.data.AUTOTUNE,
      dataset_loader: Optional[
          Callable[[xarray.Dataset], xarray.Dataset]
      ] = None,
  ):
    if dataset_loader is None:
      # In principle, it could make sense to support passing alternative
      # loaders, such as xarray_tensorstore.read() or a dask loader that calls
      # .compute(). We don't yet have any use cases where this seems to make a
      # difference, though. (The thread pool loader works as well as
      # xarray_tensorstore.read.)
      dataset_loader = _thread_pool_loader()

    if sample_dim not in source.dims:
      raise ValueError(
          'source does not include variables with a'
          f' {sample_dim!r} dimension:\n{source}'
      )
    source = _drop_static_vars(source)
    source = source.transpose(sample_dim, ...)

    block_size = _calculate_block_size(
        source,
        block_dims=[sample_dim],
        bytes_per_request=block_size_in_bytes,
        min_elements_per_request=sampler.example_size,
    )

    block_slices = sampler.list_block_slices(
        block_size, source.sizes[sample_dim]
    )

    bytes_per_element = _xarray_bytes_per_element(source, {sample_dim})
    bytes_per_example = sampler.example_size * bytes_per_element
    examples_per_block = sampler.examples_per_block(block_size)
    sample_bytes_per_block = bytes_per_example * examples_per_block
    expansion = sample_bytes_per_block / block_size_in_bytes
    logging.info(
        f'picked block_size={block_size}, corresponding to {len(block_slices)} '
        f'blocks with examples_per_block={examples_per_block}, based on '
        f'sampler={sampler} and {block_size_in_bytes=:g}. '
        f'{sample_bytes_per_block=:g} is a {expansion:1.2f}x expansion.'
    )

    self.source = source
    self.sampler = sampler
    self.block_selector = block_selector
    self.parallel_block_reads = parallel_block_reads
    self.parallel_samples = parallel_samples
    self.dataset_loader = dataset_loader

    self.block_size = block_size
    self.block_slices = block_slices
    self.bytes_per_example = bytes_per_example
    self.examples_per_block = examples_per_block

  def read(self) -> tf.data.Dataset:
    """Read this dataset into a tf.data.Dataset."""

    def generate_blocks():
      for block in self.block_selector.select(self.block_slices):
        yield (block.start, block.stop)

    def np_read_block(start: np.ndarray, stop: np.ndarray) -> list[np.ndarray]:
      selection = self.source.isel(time=slice(start, stop))
      loaded = self.dataset_loader(selection)
      arrays = [x.values for x in loaded.values()]
      return arrays

    def tf_read_block(start: tf.Tensor, stop: tf.Tensor):
      dtypes = [v.dtype for v in self.source.values()]
      shapes = [(None,) + v.shape[1:] for v in self.source.values()]
      tensors = tf.numpy_function(np_read_block, [start, stop], dtypes)
      for tensor, shape in zip(tensors, shapes):
        tensor.set_shape(shape)
      return dict(zip(self.source.keys(), tensors))

    data = tf.data.Dataset.from_generator(
        generate_blocks, output_signature=2 * (tf.TensorSpec((), tf.int64),)
    )
    data = data.map(tf_read_block, num_parallel_calls=self.parallel_block_reads)
    data = data.map(
        self.sampler.sample_block, num_parallel_calls=self.parallel_samples
    )

    data = data.unbatch()

    return data


def read_timeseries(
    source: xarray.Dataset,
    sampler: Sampler,
    block_selector: Selector = CompleteSelector(),
    *,
    sample_dim: str = 'time',
    block_size_in_bytes: float = 1e8,
    parallel_block_reads: int = tf.data.AUTOTUNE,
    parallel_samples: int = tf.data.AUTOTUNE,
) -> tf.data.Dataset:
  """Read a time-series xarray.Dataset into a tf.data.Dataset of windows.

  See go/whirl-zarr-reader for a detailed description of the design.

  Args:
    source: lazy xarray.Dataset, e.g., opened from a Zarr file with
      `open_zarr(..., chunks=None)`. All data variables with a 'time' dimension
      will be sampled. Note: setting `chunks=None` to avoid using Dask is
      preferred for optimal performance.
    sampler: specification of what time-series samples of this dataset should
      look like. Currently the only supported sampler is Windower.
    block_selector: selector called at each pass through the source dataset,
      indicating the blocks to read in order. The returned blocks should be a
      subset of passed in blocks.
    sample_dim: name of the dimension to sample along.
    block_size_in_bytes: number of bytes to use for each reading a "block" of
      data from the source data. Larger block sizes are more efficient.
    parallel_block_reads: number of blocks to read in parallel.
    parallel_samples: number of threads to use for generating samples from
      blocks.

  Returns:
    tf.data.Dataset where each element is a dict of arrays.
  """
  return _Reader(
      source=source,
      sampler=sampler,
      block_selector=block_selector,
      sample_dim=sample_dim,
      block_size_in_bytes=block_size_in_bytes,
      parallel_block_reads=parallel_block_reads,
      parallel_samples=parallel_samples,
  ).read()


def read_shuffled_shard(
    source: xarray.Dataset,
    sampler: Sampler,
    *,
    sample_dim: str = 'time',
    block_size_in_bytes: float = 1e8,
    buffer_size_in_bytes: float = 1e10,
    min_buffer_blocks: float = 10,
    parallel_block_reads: int = tf.data.AUTOTUNE,
    parallel_samples: int = tf.data.AUTOTUNE,
    shard_index: Optional[int] = None,
    shard_count: Optional[int] = None,
    seed: int = 0,
    reshuffle_each_iteration: bool = True,
) -> tf.data.Dataset:
  """Read a time-series with samples in randomly shuffled order.

  Args:
    source: lazy xarray.Dataset, e.g., opened from a Zarr file with
      `open_zarr(..., chunks=None)`. All data variables with a 'time' dimension
      will be sampled. Note: setting `chunks=None` to avoid using Dask is
      preferred for optimal performance.
    sampler: specification of what time-series samples of this dataset should
      look like.
    sample_dim: name of the dimension to sample along.
    block_size_in_bytes: number of bytes to use for each reading a "block" of
      data from the source data. Larger block sizes are more efficient.
    buffer_size_in_bytes: number of bytes to use in the shuffle buffer.
    min_buffer_blocks: minimum number of blocks that must be represented in the
      shuffle buffer, if more than one sample is taken from each block.
      Typically this should be at least as large as the batch size.
    parallel_block_reads: number of blocks to read in parallel.
    parallel_samples: number of threads to use for generating samples from
      blocks.
    shard_index: integer index for this shard of the data, in the range `[0,
      shard_count)`. In a multi-host JAX training setup, this should equal
      `jax.process_index()`.
    shard_count: total number of data shards. In a multi-host JAX training
      setup, this should equal `jax.process_count()`.
    seed: seed to use for random number generation.
    reshuffle_each_iteration: whether to use a new shuffle order for elements
      after each iteration through `source` or not.

  Returns:
    tf.data.Dataset where each element is a dict of arrays.
  """
  if shard_index is None and shard_count is None:
    shard_index = 0
    shard_count = 1

  if shard_index is None or shard_count is None:
    raise ValueError('must set both or neither of shard_index and shard_count')

  selector = ComposedSelector([
      ShardSelector(shard_index, shard_count),
      ShuffleSelector(seed, reshuffle_each_iteration),
  ])

  def _make_reader(block_size_in_bytes):
    reader = _Reader(
        source=source,
        sampler=sampler,
        sample_dim=sample_dim,
        block_selector=selector,
        block_size_in_bytes=block_size_in_bytes,
        parallel_block_reads=parallel_block_reads,
        parallel_samples=parallel_samples,
    )
    buffer_size = int(buffer_size_in_bytes / reader.bytes_per_example)
    logging.info(
        f'picked shuffle buffer size of {buffer_size} based on '
        f'{buffer_size_in_bytes=:g}'
    )
    return reader, buffer_size

  reader, buffer_size = _make_reader(block_size_in_bytes)

  if buffer_size:
    examples_per_block = reader.examples_per_block
    buffer_blocks = buffer_size / examples_per_block
    if examples_per_block > 1 and buffer_blocks < min_buffer_blocks:
      block_size_in_bytes = reader.bytes_per_example
      logging.warning(
          'insufficient diversity in proposed shuffle buffer: '
          f'{examples_per_block=} and {buffer_size=} means that on average '
          f'only {buffer_blocks:g} blocks will be represented in the shuffle '
          f'buffer, which is less than {min_buffer_blocks=}. Falling back to '
          f'one example per block ({block_size_in_bytes=:g}).'
      )
      reader, buffer_size = _make_reader(block_size_in_bytes)
      assert reader.examples_per_block == 1

  data = reader.read()

  if buffer_size:
    # for testing, disable the shuffle buffer if it has size zero (the shuffle
    # method does not support size zero buffers)
    data = data.shuffle(buffer_size, seed, reshuffle_each_iteration)

  return data



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LinearTransforms for use in Metrics."""
import dataclasses
import functools
from typing import Callable, Mapping, Optional, Sequence
from dinosaur import coordinate_systems
from dinosaur import filtering
from dinosaur import horizontal_interpolation
from dinosaur import pytree_utils
from dinosaur import spherical_harmonic
from dinosaur import typing
import gin
import jax
import jax.numpy as jnp
import numpy as np

import metrics_util


Pytree = typing.Pytree
TrajectoryRepresentations = typing.TrajectoryRepresentations

tree_leaves = jax.tree_util.tree_leaves
tree_map = jax.tree_util.tree_map


@dataclasses.dataclass
class LinearTransform:
  """A linear transformation, for TransformedL2Loss."""

  trajectory_spec: metrics_util.TrajectorySpec

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    raise NotImplementedError


LinearTransformConstructor = Callable[
    [metrics_util.TrajectorySpec], LinearTransform
]


@dataclasses.dataclass
class ComposedTransformForLoss(LinearTransform):
  """Composition of multiple linear transformations for computation of loss.

  Attributes:
    components: components[i](self.trajectory_spec) initializes the i + 1 member
      of self.transforms.
    transforms: errors are transformed as error --> transforms[0](error) -->
      transforms[1](error) --> ⋯. The 0th transform is inserted by this class as
      TruncateToTrajectoryLength.
  """

  components: Sequence[LinearTransformConstructor]
  transforms: Sequence[LinearTransform] = dataclasses.field(init=False)

  def __post_init__(self):
    # Insert TruncateToTrajectoryLength first in all cases. It's okay if it was
    # already inserted... it is idempotent. This ensures that
    #   len(self.transforms) = len(self.components) + 1
    # in all cases.
    components = [TruncateToTrajectoryLength] + list(self.components)
    self.transforms = [
        constructor(self.trajectory_spec) for constructor in components
    ]

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    for transform in self.transforms:
      errors = transform(errors, targets)
    return errors


@gin.register
@dataclasses.dataclass
class LegacyTimeRescaling(LinearTransform):
  """Time scaling from WeightedL2CumulativeLoss."""

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    n = self.trajectory_spec.trajectory_length
    steps_per_save = self.trajectory_spec.steps_per_save
    scale = 1 if n == 1 else 1 / np.sqrt((n - 1) * steps_per_save)
    return tree_map(lambda x: x * scale, errors)


@gin.register
@dataclasses.dataclass
class TimeRescaling(LinearTransform):
  """Time scaling that assumes error grows like a random walk.

  This rescales errors like
    errors --> errors / σ(T),
    σ(T) := sqrt( sum(variance) / variance(T) )
  where variance(T) is the assumed variance. A random walk has variance ∝ T.
  This function uses similar scaling.

  See also:
  * Climatology vs. ENS CRPS values indicate skill difficult after 240 hrs
    http://screen/8sVodqThEk6o693
  * Plotting this function for various parameter values
    http://screen/AubXNomsgm7g92o and http://gpaste/6727081386835968

  Attributes:
    base_squared_error_in_hours: Number of hours before assumed variance starts
      growing (almost) linearly.
    asymptotic_squared_error_in_hours: Number of hours before assumed variance
      slows its growth. Set to None (the default) if variance grows indefinitely
  """

  base_squared_error_in_hours: float
  asymptotic_squared_error_in_hours: Optional[float] = None

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    time_sizes = np.unique([x.shape[0] for x in tree_leaves(errors)])
    if time_sizes.size != 1:
      raise ValueError(f'Expected unique time dimension size. {time_sizes=}')
    time_size = time_sizes[0]
    if self.trajectory_spec.trajectory_length != time_size:
      logging.info(
          f'errors has {time_size=} !='
          f' {self.trajectory_spec.trajectory_length=}. This is probably due to'
          ' the Loss slicing via the time_step kwarg. Will use {time_size=}'
          ' to compute scaling.'
      )

    steps_per_save = self.trajectory_spec.steps_per_save
    t = np.arange(time_size) * steps_per_save
    if self.asymptotic_squared_error_in_hours is not None:
      # Rescale "time" `t`, so it stops growing when
      #   t >> asymptotic_squared_error_in_hours.
      t = t / (1 + t / self.asymptotic_squared_error_in_hours)

    inv_variance = 1 / (1 + t / self.base_squared_error_in_hours)
    scale = np.sqrt(inv_variance / inv_variance.sum())
    scale = scale.reshape(-1, 1, 1, 1)

    return tree_map(lambda x: x * scale, errors)


@gin.register
@dataclasses.dataclass
class CustomTimeRescaling(LinearTransform):
  """Custom time scaling that uses pre-specified values."""

  scaling_weights: Sequence[float]

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    n = self.trajectory_spec.trajectory_length
    scale = np.asarray(self.scaling_weights)[:n].reshape(-1, 1, 1, 1)
    return tree_map(lambda x: x * scale, errors)


@gin.register
@dataclasses.dataclass
class DelayedTimeRescaling(LinearTransform):
  """Time scaling with smooth delay that transitions into hyperbolic decay."""

  base_squared_error_in_hours: float
  delay_power: float = 1.0
  decay_power: float = 1.0

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    n = self.trajectory_spec.trajectory_length
    steps_per_save = self.trajectory_spec.steps_per_save
    t = np.arange(n) * steps_per_save

    a = 1 / self.base_squared_error_in_hours
    inv_variance = 1 / (
        (1 + (a * t) ** self.delay_power) ** (1/self.decay_power))
    scale = np.sqrt(inv_variance / inv_variance.sum())
    scale = scale.reshape(-1, 1, 1, 1)

    return tree_map(lambda x: x * scale, errors)


@gin.register
@dataclasses.dataclass
class TruncateToTrajectoryLength(LinearTransform):
  """Truncate errors to self.trajectory_spec.trajectory_length.

  To ensure loss is computed over the correct trajectory length, this transform
  should be used as the first step in any ComposedTransformForLoss.
  """

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    n = self.trajectory_spec.trajectory_length
    return metrics_util.extract_time_slice(errors, slice(0, n))


@gin.register
@dataclasses.dataclass
class TotalWavenumberMasking(LinearTransform):
  """Transform that masks out wavenumbers greater than `max_wavenumber`."""

  max_wavenumber: int
  is_encoded: bool = False

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    if self.is_encoded:
      grid = self.trajectory_spec.coords.horizontal
    else:
      grid = self.trajectory_spec.data_coords.horizontal

    modal_shape = grid.modal_shape
    mask = np.arange(modal_shape[-1]) < self.max_wavenumber
    mask = mask.astype(float)
    return tree_map(lambda x: x * mask, errors)


@gin.register
@dataclasses.dataclass
class ConservativeRegridder(LinearTransform):
  """Linear transform that regrids."""

  def __init__(
      self,
      trajectory_spec: metrics_util.TrajectorySpec,
      target_grid: spherical_harmonic.Grid,
  ):
    super().__init__(trajectory_spec=trajectory_spec)
    self.regridder = horizontal_interpolation.ConservativeRegridder(
        source_grid=trajectory_spec.coords.horizontal, target_grid=target_grid
    )

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # Unused
    return tree_map(self.regridder, errors)


@gin.register
@dataclasses.dataclass
class PerVariableRescaling(LinearTransform):
  """Transform that reweights contribution per variable."""
  weights: Pytree
  scale: float = 1.0

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    weights = self.weights
    if weights is None:
      weights = tree_map(lambda x: 1.0, errors)
    else:
      weights = pytree_utils.replace_with_matching_or_default(
          errors, weights, default=None,
          check_used_all_replace_keys=True,
      )
    root_weights = tree_map(lambda w: np.sqrt(w * self.scale), weights)
    return tree_map(jnp.multiply, errors, root_weights)


@gin.register
class ExponentialFilteringByLeadtime(LinearTransform):
  """Applied leadtime dependent exponential filters to errors."""

  def __init__(
      self,
      trajectory_spec: metrics_util.TrajectorySpec,
      filter_attenuations: typing.Pytree,
      filter_orders: typing.Pytree,
      is_encoded: bool = False,
  ):
    super().__init__(trajectory_spec=trajectory_spec)
    n = trajectory_spec.trajectory_length
    if is_encoded:
      grid = trajectory_spec.coords.horizontal
    else:
      grid = trajectory_spec.data_coords.horizontal
    # expand dims for `level, lon, total wavenumbers` so that filter parameters
    # are applied to different time values.
    to_array_fn = lambda x: np.expand_dims(np.array(x)[:n], axis=(1, 2, 3))
    is_leaf = lambda x: isinstance(x, Sequence)
    attenuations = tree_map(to_array_fn, filter_attenuations, is_leaf=is_leaf)
    orders = tree_map(to_array_fn, filter_orders, is_leaf=is_leaf)
    self.filter_fns = tree_map(
        lambda a, p: filtering.exponential_filter(grid, a, p),
        attenuations,
        orders,
    )

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    filter_fns = pytree_utils.replace_with_matching_or_default(
        errors, self.filter_fns, default=None, check_used_all_replace_keys=True)
    return tree_map(lambda fn, err: fn(err), filter_fns, errors)


@gin.register
class LevelRescaling(LinearTransform):
  """Linear transform that scales values with vertical levels."""

  def __init__(
      self,
      trajectory_spec: metrics_util.TrajectorySpec,
      scale: Sequence[float],
      keys_to_scale: Sequence[str] = tuple(),
  ):
    super().__init__(trajectory_spec)
    self.scale_fn = functools.partial(
        coordinate_systems.scale_levels_for_matching_keys,
        scales=np.asarray(scale),
        keys_to_scale=keys_to_scale,
    )

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    return self.scale_fn(errors)


@gin.register
class LevelRemoval(LinearTransform):
  """Linear transform that removes vertical levels."""

  def __init__(
      self,
      trajectory_spec: metrics_util.TrajectorySpec,
      keep_levels: Sequence[float],
  ):
    super().__init__(trajectory_spec)
    n_levels = trajectory_spec.data_coords.vertical.layers
    indices = jnp.array([i for i in range(n_levels) if keep_levels[i]])
    self.take_arr = lambda x: jnp.take(x, indices, axis=metrics_util.LEVEL_AXIS)

  def __call__(self, errors: Pytree, targets: Pytree) -> Pytree:
    del targets  # unused.
    return tree_map(self.take_arr, errors)



"""Testing utilities for coordax."""

from typing import Mapping

import chex
import jax
from neuralgcm.experimental.coordax import core
import numpy as np


AxisName = core.AxisName


def assert_field_properties(
    actual: core.Field,
    data: np.ndarray | jax.Array | None = None,
    dims: tuple[AxisName, ...] | None = None,
    shape: tuple[int, ...] | None = None,
    coord_field_keys: set[AxisName] | None = None,
    named_shape: Mapping[AxisName, int] | None = None,
    positional_shape: tuple[int, ...] | None = None,
    rtol: float = 1e-5,
    atol: float = 1e-5,
):
  """Asserts that a Field has expected properties."""
  if data is not None:
    np.testing.assert_allclose(actual.data, data, rtol=rtol, atol=atol)
  if dims is not None:
    chex.assert_equal(actual.dims, dims)
  if shape is not None:
    chex.assert_shape(actual, shape)
  if coord_field_keys is not None:
    chex.assert_trees_all_equal(
        set(actual.coord_fields.keys()), coord_field_keys
    )
  if named_shape is not None:
    chex.assert_equal(actual.named_shape, named_shape)
  if positional_shape is not None:
    chex.assert_equal(actual.positional_shape, positional_shape)


def assert_fields_allclose(
    actual: core.Field,
    desired: core.Field,
    rtol: float = 1e-5,
    atol: float = 1e-5,
):
  """Asserts that two Fields are close and have matching coordinates."""
  assert_field_properties(
      actual=actual,
      data=desired.data,
      dims=desired.dims,
      shape=desired.shape,
      coord_field_keys=set(desired.coord_fields.keys()),
      named_shape=desired.named_shape,
      positional_shape=desired.positional_shape,
      rtol=rtol,
      atol=atol,
  )



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ForecastSystem API."""

from __future__ import annotations

import abc
import dataclasses
from typing import Any, Callable

from etils import epath
import fiddle as fdl
from flax import nnx
import jax
from neuralgcm.experimental import checkpointing  # pylint: disable=unused-import
from neuralgcm.experimental import coordax as cx
from neuralgcm.experimental import coordinates
from neuralgcm.experimental import diagnostics
from neuralgcm.experimental import dynamic_io
from neuralgcm.experimental import module_utils
from neuralgcm.experimental import random_processes
from neuralgcm.experimental import time_integrators
from neuralgcm.experimental import typing
import numpy as np
import orbax.checkpoint as ocp
import pandas as pd
import xarray


def calculate_sub_steps(
    timestep: np.timedelta64, duration: typing.TimedeltaLike
) -> int:
  """Calculate the number of time-steps required to simulate a time interval."""
  duration = pd.Timedelta(duration)
  time_step_ratio = duration / timestep
  if abs(time_step_ratio - round(time_step_ratio)) > 1e-6:
    raise ValueError(
        f'non-integral time-step ratio: {duration=} is not a multiple of '
        f'the internal model timestep {timestep}'
    )
  return round(time_step_ratio)


@dataclasses.dataclass
class ForecastSystem(nnx.Module, abc.ABC):
  """Base class for forecast systems."""

  _metadata: dict = dataclasses.field(default_factory=dict, kw_only=True)  # pylint: disable=g-bare-generic

  @property
  def metadata(self):
    """Returns optional metadata associated with the model."""
    return self._metadata

  @property
  def timestep(self):
    """Returns the timestep of the model."""
    raise NotImplementedError()

  def update_metadata(self, key, value):
    """Adds metadata to the model."""
    self._metadata[key] = value

  @abc.abstractmethod
  def assimilate_prognostics(
      self,
      observations: typing.ObservationData,
      dynamic_inputs: typing.Pytree | None = None,
      rng: typing.PRNGKeyArray | None = None,
      initial_state: typing.ModelState | None = None,
  ) -> typing.Pytree:
    raise NotImplementedError()

  @abc.abstractmethod
  def advance_prognostics(
      self, prognostics: typing.PyTreeState
  ) -> typing.PyTreeState:
    raise NotImplementedError()

  @abc.abstractmethod
  def observe_from_prognostics(
      self,
      prognostics: typing.Pytree,
      query: typing.ObservationSpecs,
      dynamic_inputs: typing.Pytree | None = None,
  ) -> typing.Pytree:
    raise NotImplementedError()

  def assimilate(
      self,
      observations: typing.ObservationData,
      dynamic_inputs: typing.Pytree | None = None,
      rng: typing.PRNGKeyArray | None = None,
      initial_state: typing.ModelState | None = None,
  ) -> typing.ModelState:
    self.update_dynamic_inputs(dynamic_inputs)
    self.initialize_random_processes(rng)
    prognostics = self.assimilate_prognostics(observations, initial_state)
    diagnostic = nnx.state(self, diagnostics.DiagnosticValue)
    randomness = nnx.state(self, random_processes.RandomnessValue)
    return typing.ModelState(prognostics, diagnostic, randomness)

  def advance(
      self,
      state: typing.ModelState,
      dynamic_inputs: typing.Pytree | None = None,
  ) -> typing.ModelState:
    self.update_dynamic_inputs(dynamic_inputs)
    nnx.update(self, state.diagnostics)
    nnx.update(self, state.randomness)
    prognostics = self.advance_prognostics(state.prognostics)
    diagnostic = nnx.state(self, diagnostics.DiagnosticValue)
    randomness = nnx.state(self, random_processes.RandomnessValue)
    return typing.ModelState(prognostics, diagnostic, randomness)

  def observe(
      self,
      state: typing.ModelState,
      query: typing.ObservationSpecs,
      dynamic_inputs: typing.Pytree | None = None,
  ) -> typing.Pytree:
    self.update_dynamic_inputs(dynamic_inputs)
    nnx.update(self, state.diagnostics)
    nnx.update(self, state.randomness)
    return self.observe_from_prognostics(state.prognostics, query)

  def update_dynamic_inputs(self, dynamic_input: typing.Pytree | None = None):
    if dynamic_input is not None:
      for covariate_module in module_utils.retrieve_subclass_modules(
          self, dynamic_io.DynamicInputModule
      ):
        covariate_module.update_dynamic_inputs(dynamic_input)

  def initialize_random_processes(self, rng: typing.PRNGKeyArray) -> None:
    modules = module_utils.retrieve_subclass_modules(
        self, random_processes.RandomProcessModule
    )
    rngs = jax.random.split(rng, len(modules))
    for random_process, key in zip(modules, rngs):
      random_process.unconditional_sample(key)

  def format_diagnostics(
      self, state: typing.ModelState | None = None, time: float | None = None
  ) -> typing.Pytree:
    if state is not None:
      nnx.update(self, state.diagnostics)
    outputs = {}
    for diagnostic_module in module_utils.retrieve_subclass_modules(
        self, diagnostics.DiagnosticModule
    ):
      outputs |= diagnostic_module.format_diagnostics(time)
    return outputs

  def unroll(
      self,
      state: typing.ModelState,
      outer_steps: int,
      timedelta: typing.TimedeltaLike | None = None,
      start_with_input: bool = True,
      post_process_fn: Callable[..., typing.Pytree] = lambda x, **kwargs: x,
      dynamic_inputs: typing.Pytree | None = None,
  ) -> tuple[typing.ModelState, typing.Pytree]:
    self.update_dynamic_inputs(dynamic_inputs)
    nnx.update(self, state.diagnostics)
    nnx.update(self, state.randomness)

    if timedelta is None:
      timedelta = self.timestep
    inner_steps = calculate_sub_steps(self.timestep, timedelta)

    def _inner_step(model_and_prognostics):
      model, prognostics = model_and_prognostics
      next_prognostics = model.advance_prognostics(prognostics)
      return (model, next_prognostics)

    inner_step = time_integrators.repeated(_inner_step, inner_steps)

    def _step(model_and_state):
      model, model_state = model_and_state
      model, next_prognostics = inner_step((model, model_state.prognostics))
      diagnostic = nnx.state(model, diagnostics.DiagnosticValue)
      randomness = nnx.state(model, random_processes.RandomnessValue)
      next_model_state = typing.ModelState(
          next_prognostics, diagnostic, randomness
      )
      frame = model_state if start_with_input else next_model_state
      return (model, next_model_state), post_process_fn(frame, model=model)

    unroll_fn = nnx.scan(
        _step,
        length=outer_steps,
        in_axes=nnx.Carry,
        out_axes=(nnx.Carry, 0),
    )
    (_, final_state), intermediates = unroll_fn((self, state))
    time = coordinates.TimeDelta(np.arange(outer_steps) * inner_steps)
    tag_prefix = lambda x: x.tag_prefix(time) if isinstance(x, cx.Field) else x
    intermediates = jax.tree.map(tag_prefix, intermediates)
    return final_state, intermediates

  def inputs_from_xarray(self, ds: xarray.Dataset) -> typing.Pytree:
    """Converts xarray dataset to inputs for the model."""
    raise NotImplementedError(
        f'Class {self.__class__.__name__} does not implement'
        ' inputs_from_xarray.'
    )

  def dynamic_inputs_from_xarray(self, ds: xarray.Dataset) -> typing.Pytree:
    """Converts xarray dataset to dynamic covariates for the model."""
    raise NotImplementedError(
        f'Class {self.__class__.__name__} does not implement'
        ' dynamic_inputs_from_xarray.'
    )

  def data_from_xarray(self, ds: xarray.Dataset) -> typing.Pytree:
    """Converts xarray dataset to data for the model."""
    return self.inputs_from_xarray(ds), self.dynamic_inputs_from_xarray(ds)

  @classmethod
  def checkpoint_args(cls, mode: str = 'restore'):
    """Returns orbax checkpoint args for the model for mode=`mode`."""
    if mode == 'restore':
      arg = ocp.args.PyTreeRestore
    else:
      arg = ocp.args.PyTreeSave
    # TODO(dkochkov) Consider supporting partitioned params.
    return {
        'params': arg,
        'metadata': arg,
    }

  @classmethod
  def from_checkpoint(
      cls,
      path: str | epath.PathLike,
      checkpointer: ocp.Checkpointer | None = None,
  ) -> tuple[ForecastSystem, nnx.State]:
    checkpoint_args = cls.checkpoint_args('restore')
    if checkpointer is None:
      handler = ocp.CompositeCheckpointHandler(*checkpoint_args.keys())
      checkpointer = ocp.Checkpointer(handler)
    restored = checkpointer.restore(
        path, args=ocp.args.Composite(metadata=checkpoint_args.pop('metadata'))
    )
    metadata = restored['metadata']
    if 'fiddle_config' in metadata:
      cfg = metadata['fiddle_config']
      # Note: nnx.eval_shape doesn't seem to work here because numerics
      # components might need to make use of init values beyond shape.
      model = fdl.build(cfg)
      assert isinstance(model, ForecastSystem)  # make pytype happy.
      model.update_metadata('fiddle_config', cfg)
      params_structure = nnx.state(model, nnx.Variable)
      params = checkpointer.restore(
          path,
          args=ocp.args.Composite(
              params=checkpoint_args['params'](params_structure)
          ),
      )['params']
      nnx.update(model, params)
      return model, params
    else:
      # TODO(dkochkov) implement restoration from pickled params and graph_def.
      raise ValueError('Not yet supported without fiddle_config.')

  def checkpoint_state(
      self,
      mode: str = 'restore',
      param_types_filter: Any = nnx.Variable,
  ):
    # TODO(dkochkov) support different param_types_filter.
    # This likely requires serialization of the filter sequence in a separate
    # field and then recreating the state split in `from_checkpoint` method.
    if param_types_filter != nnx.Variable:
      raise ValueError(f'{param_types_filter=}!=nnx.Variable.')
    params_state = nnx.state(self, param_types_filter)
    metadata = self.metadata
    ckpt_items = {
        'params': params_state,
        'metadata': metadata,
    }
    ckpt_args = self.checkpoint_args(mode)
    return {k: arg(ckpt_items[k]) for k, arg in ckpt_args.items()}



# Copyright 2024 Google LLC

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper utilities for checkpointing Fiddle configs used in NeuralGCM."""

import asyncio
from concurrent import futures
import functools
from typing import Any, Optional, Sequence

from etils import epath
import fiddle as fdl
from fiddle.experimental import serialization
import orbax.checkpoint as ocp


ParamInfo = ocp.type_handlers.ParamInfo
Metadata = ocp.metadata.value.Metadata


class FiddleConfigHandler(ocp.type_handlers.TypeHandler):
  """A wrapper around serialization of fdl.Config to the json format."""

  def __init__(
      self, filename: str | None = None, primary_host: int | None = 0
  ):
    """Initializes FiddleConfigHandler.

    Args:
      filename: optional file name given to the written file; defaults to
        'fiddle_config'
      primary_host: the host id of the primary host.  Default to 0.  If it's set
        to None, then all hosts will be considered as primary.  It's useful in
        the case that all hosts are only working with local storage.
    """
    self._filename = filename or 'fiddle_config'
    self._primary_host = primary_host
    self._executor = futures.ThreadPoolExecutor(max_workers=1)

  def _save_fn(self, x, directory):
    if ocp.utils.is_primary_host(self._primary_host):
      path = directory / self._filename
      path.write_text(x)
    return 0

  def typestr(self) -> str:
    return 'Config'

  async def serialize(
      self,
      values: Sequence[Any],
      infos: Sequence[ParamInfo],
      args: Sequence[ocp.SaveArgs] | None = None,
  ) -> list[futures.Future]:  # pylint: disable=g-bare-generic
    """Serializes the given fdl.Config to the json format."""
    del args  # Unused in this example.
    ocp_futures = []
    for value, info in zip(values, infos):
      # make sure the per-key directory is present as OCDBT doesn't create one
      info.path.mkdir(exist_ok=True)  # pylint: disable=attribute-error
      ocp_futures.append(
          self._executor.submit(
              functools.partial(self._write_config, value, info.path)
          )
      )
    return ocp_futures

  async def deserialize(
      self,
      infos: Sequence[ParamInfo],
      args: Optional[Sequence[ocp.RestoreArgs]] = None,
  ):
    del args  # Unused in this example.
    ocp_futures = []
    for info in infos:
      ocp_futures.append(
          await asyncio.get_event_loop().run_in_executor(
              self._executor,
              functools.partial(self._from_serialized, info.path),
          )
      )
    return await asyncio.gather(*ocp_futures)

  async def metadata(self, infos: Sequence[ParamInfo]) -> Sequence[Metadata]:
    return [Metadata(name=info.name, directory=info.path) for info in infos]

  def _write_config(self, config: fdl.Config, path: epath.Path):
    serialized_conifig = serialization.dump_json(config, indent=4)
    self._save_fn(serialized_conifig, path)
    return path

  async def _from_serialized(self, path: epath.Path):
    path = path / self._filename
    serialized_config = path.read_text()
    return serialization.load_json(serialized_config)


ocp.type_handlers.register_type_handler(
    fdl.Config, FiddleConfigHandler(), override=True
)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for towers.py."""

from absl.testing import absltest
from absl.testing import parameterized
import gin
import haiku as hk
import jax
from neuralgcm import towers  # pylint: disable=unused-import
import numpy as np


class ColumnTowerTest(parameterized.TestCase):
  """Tests Column tower."""

  @parameterized.parameters(
      dict(
          output_size=2,
          num_hidden_units=3,
          num_hidden_layers=4,
          with_bias=True),
      dict(
          output_size=4,
          num_hidden_units=5,
          num_hidden_layers=6,
          with_bias=False),
      )
  def test_column_tower_mlp_shapes(
      self,
      output_size: int,
      num_hidden_units: int,
      num_hidden_layers: int,
      with_bias: bool,
  ):
    gin.enter_interactive_mode()

    @gin.configurable
    def tower_forward(x, *, tower):  # defines a forward function
      return tower(output_size)(x)

    config = '\n'.join([
        'tower_forward.tower = @ColumnTower',
        'ColumnTower.column_net_factory = @MlpUniform',
        'ColumnTower.name = "column_tower"',
        f'MlpUniform.num_hidden_units = {num_hidden_units}',
        f'MlpUniform.num_hidden_layers = {num_hidden_layers}',
        f'MlpUniform.with_bias = {with_bias}',
    ])

    gin.clear_config()
    gin.parse_config(config)

    n_features = 8
    input_size = (n_features, 10, 20)
    inputs = np.random.uniform(size=input_size)
    model = hk.without_apply_rng(hk.transform(tower_forward))
    params = model.init(jax.random.PRNGKey(42), inputs)
    out = model.apply(params, inputs)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_size, 10, 20))
    with self.subTest('total_params_count'):
      expected = count_mlp_uniform_params(
          input_size=n_features,
          num_hidden_units=num_hidden_units,
          num_hidden_layers=num_hidden_layers,
          output_size=output_size,
          with_bias=with_bias)
      actual = sum([np.prod(x.shape)
                    for x in jax.tree_util.tree_leaves(params)])
      self.assertEqual(actual, expected)


class VerticalConvTowerTest(parameterized.TestCase):
  """Tests VerticalConvTower tower."""

  @parameterized.parameters(
      dict(
          output_size=7,
          channels=[4],
          kernel_shape=3,
      ),
      dict(
          output_size=20,
          channels=[4, 8, 10],
          kernel_shape=5,
      ),
  )
  def test_conv_tower_shapes(
      self,
      output_size: int,
      channels: list[int],
      kernel_shape: int,
  ):
    """Tests that VerticalConvTower outputs and params have expected shapes."""
    gin.enter_interactive_mode()

    @gin.configurable
    def tower_forward(x, tower):  # defines a forward function
      return tower(output_size)(x)

    config = '\n'.join([
        'tower_forward.tower = @VerticalConvTower',
        f'VerticalConvTower.channels = {channels}',
        f'VerticalConvTower.kernel_shape = {kernel_shape}',
        'VerticalConvTower.name = "conv1D_tower"',
    ])

    gin.clear_config()
    gin.parse_config(config)
    n_levels = 24
    in_channels = 8
    input_shape = (in_channels, n_levels, 10, 20)
    key = jax.random.PRNGKey(42)
    inputs = jax.random.uniform(key, shape=input_shape)
    model = hk.without_apply_rng(hk.transform(tower_forward))
    params = model.init(key, inputs)
    out = model.apply(params, inputs)
    tot_channels = channels
    tot_channels.append(output_size)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_size, n_levels, 10, 20))
    with self.subTest('total_params_count'):
      expected = count_conv1d_tower_params(
          in_channels=in_channels,
          kernel_shape=kernel_shape,
          channels=tot_channels,
          )
      actual = sum([np.prod(x.shape)
                    for x in jax.tree_util.tree_leaves(params)])
      self.assertEqual(actual, expected)


class Conv2DTowerTest(parameterized.TestCase):
  """Tests ConvNet tower."""

  @parameterized.parameters(
      dict(
          output_size=2,
          num_hidden_units=3,
          num_hidden_layers=4,
          kernel_shape=(3, 3),
          with_bias=True,),
      dict(
          output_size=4,
          num_hidden_units=5,
          num_hidden_layers=6,
          kernel_shape=(5, 5),
          with_bias=False,),
      )
  def test_conv_tower_shapes(
      self,
      output_size: int,
      num_hidden_units: int,
      num_hidden_layers: int,
      kernel_shape: tuple[int, int],
      with_bias: bool,
  ):
    """Tests that Conv2DTower outputs and params have expected shapes."""
    gin.enter_interactive_mode()

    @gin.configurable
    def tower_forward(x, tower):  # defines a forward function
      return tower(output_size)(x)

    config = '\n'.join([
        'tower_forward.tower = @Conv2DTower',
        f'Conv2DTower.num_hidden_units = {num_hidden_units}',
        f'Conv2DTower.num_hidden_layers = {num_hidden_layers}',
        f'Conv2DTower.kernel_shape = {kernel_shape}',
        f'Conv2DTower.with_bias = {with_bias}',
        'Conv2DTower.name = "conv_tower"',
    ])

    gin.clear_config()
    gin.parse_config(config)

    n_features = 8
    input_size = (n_features, 10, 20)
    inputs = np.random.uniform(size=input_size)
    model = hk.without_apply_rng(hk.transform(tower_forward))
    params = model.init(jax.random.PRNGKey(42), inputs)
    out = model.apply(params, inputs)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_size, 10, 20))
    with self.subTest('total_params_count'):
      expected = count_conv_tower_params(
          input_size=n_features,
          kernel_shape=kernel_shape,
          num_hidden_units=num_hidden_units,
          num_hidden_layers=num_hidden_layers,
          output_size=output_size,
          with_bias=with_bias)
      actual = sum([np.prod(x.shape)
                    for x in jax.tree_util.tree_leaves(params)])
      self.assertEqual(actual, expected)


class EpdTowerTest(parameterized.TestCase):
  """Tests encode/process/decode tower."""

  @parameterized.parameters(
      dict(
          output_size=3,
          latent_size=6,
          layer_size=7,
          num_process_layers=2,
          num_process_blocks=2,),
      dict(
          output_size=5,
          latent_size=13,
          layer_size=4,
          num_process_layers=3,
          num_process_blocks=1,),
  )
  def test_epd_tower_mlp_shapes(
      self,
      output_size: int,
      latent_size: int,
      layer_size: int,
      num_process_layers: int,
      num_process_blocks: int,
  ):
    """Tests that EpdTower outputs and params have expected shapes."""
    gin.enter_interactive_mode()

    @gin.configurable
    def tower_forward(x, tower):  # defines a forward function
      return tower(output_size)(x)

    config = '\n'.join([
        'tower_forward.tower = @EpdTower',
        f'EpdTower.latent_size = {latent_size}',
        f'EpdTower.num_process_blocks = {num_process_blocks}',
        'EpdTower.encode_tower_factory = @encode/ColumnTower',
        'EpdTower.process_tower_factory = @process/ColumnTower',
        'EpdTower.decode_tower_factory = @decode/ColumnTower',
        'encode/ColumnTower.column_net_factory = @encode/MlpUniform',
        'encode/ColumnTower.name = "encode_tower"',
        f'encode/MlpUniform.num_hidden_units = {layer_size}',
        'encode/MlpUniform.num_hidden_layers = 1',
        'encode/MlpUniform.name = "encode_layer"',
        'process/ColumnTower.column_net_factory = @process/MlpUniform',
        'process/ColumnTower.name = "process_tower"',
        f'process/MlpUniform.num_hidden_units = {layer_size}',
        f'process/MlpUniform.num_hidden_layers = {num_process_layers}',
        'process/MlpUniform.name = "process_layer"',
        'decode/ColumnTower.column_net_factory = @decode/MlpUniform',
        'decode/ColumnTower.name = "decode_tower"',
        f'decode/MlpUniform.num_hidden_units = {layer_size}',
        'decode/MlpUniform.num_hidden_layers = 2',
        'decode/MlpUniform.name = "decode_layer"',
    ])

    gin.clear_config()
    gin.parse_config(config)

    n_features = 8
    input_size = (n_features, 10, 20)
    inputs = np.random.uniform(size=input_size)
    model = hk.without_apply_rng(hk.transform(tower_forward))
    params = model.init(jax.random.PRNGKey(42), inputs)
    out = model.apply(params, inputs)
    with self.subTest('output_shape'):
      self.assertEqual(out.shape, (output_size, 10, 20))
    with self.subTest('total_params_count'):
      num_encode_params = count_mlp_uniform_params(
          input_size=n_features,
          num_hidden_units=layer_size,
          num_hidden_layers=1,
          output_size=latent_size,
          with_bias=True)
      num_process_params = count_mlp_uniform_params(
          input_size=latent_size,
          num_hidden_units=layer_size,
          num_hidden_layers=num_process_layers,
          output_size=latent_size,
          with_bias=True)
      num_decode_params = count_mlp_uniform_params(
          input_size=latent_size,
          num_hidden_units=layer_size,
          num_hidden_layers=2,
          output_size=output_size,
          with_bias=True)
      expected = (
          num_encode_params + num_process_blocks * num_process_params +
          num_decode_params)
      actual = sum([np.prod(x.shape)
                    for x in jax.tree_util.tree_leaves(params)])
      self.assertEqual(actual, expected)


def count_mlp_uniform_params(
    input_size: int,
    num_hidden_units: int,
    num_hidden_layers: int,
    output_size: int,
    with_bias: bool):
  """Returns the number of parameters in MlpUniform based on settings."""
  n_input = input_size * num_hidden_units + num_hidden_units * with_bias
  n_hidden = num_hidden_units * num_hidden_units + num_hidden_units * with_bias
  n_output = num_hidden_units * output_size + output_size * with_bias
  return n_input + n_hidden * (num_hidden_layers - 1) + n_output


def count_conv_lon_lat_params(
    input_size: int,
    kernel_shape: int,
    output_size: int,
    with_bias: bool):
  """Returns the number of parameters in ConvLonLat based on settings."""
  n_w_params = np.prod(kernel_shape) * input_size * output_size
  n_b_params = output_size
  return n_w_params + n_b_params * with_bias


def count_conv_tower_params(
    input_size: int,
    kernel_shape: int,
    num_hidden_units: int,
    num_hidden_layers: int,
    output_size: int,
    with_bias: bool):
  """Returns the number of parameters in Conv2DTower based on settings."""
  n_input = count_conv_lon_lat_params(
      input_size=input_size,
      kernel_shape=kernel_shape,
      output_size=num_hidden_units,
      with_bias=with_bias)
  n_hidden = count_conv_lon_lat_params(
      input_size=num_hidden_units,
      kernel_shape=kernel_shape,
      output_size=num_hidden_units,
      with_bias=with_bias)
  n_output = count_conv_lon_lat_params(
      input_size=num_hidden_units,
      kernel_shape=kernel_shape,
      output_size=output_size,
      with_bias=with_bias)
  return n_input + n_hidden * (num_hidden_layers - 1) + n_output


def count_conv1d_tower_params(
    in_channels: int,
    kernel_shape: int,
    channels: list[int],
    with_bias: bool = True,
    ) -> int:
  """Returns the number of parameters in Conv1D."""
  n_params = []
  input_current = in_channels
  for channel in channels:
    n_params.append(count_conv_lon_lat_params(
        input_size=input_current,
        kernel_shape=kernel_shape,
        output_size=channel,
        with_bias=with_bias))
    input_current = channel
  return sum(n_params)


if __name__ == '__main__':
  jax.config.parse_flags_with_absl()
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Physics parameterization modules that compute non-dynamical tendencies."""

from typing import Any, Callable, Optional
from dinosaur import coordinate_systems
from dinosaur import pytree_utils
from dinosaur import typing
import gin
import haiku as hk
import jax
from neuralgcm import features
from neuralgcm import mappings
from neuralgcm import transforms

FeaturesModule = features.FeaturesModule
Forcing = typing.Forcing
MappingModule = mappings.MappingModule
StepFilterModule = Callable[..., typing.PyTreeStepFilterFn]
TransformModule = typing.TransformModule


@gin.register
class DirectNeuralParameterization(hk.Module):
  """Computes modal physics tendencies from the input state and forcing."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      modal_to_nodal_features_module: FeaturesModule,
      nodal_mapping_module: mappings.MappingModule,
      tendency_transform_module: TransformModule,
      prediction_mask: Optional[typing.Pytree] = None,
      filter_module: Optional[StepFilterModule] = None,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.coords = coords
    self.prediction_mask = prediction_mask
    self.modal_to_nodal_features_fn = modal_to_nodal_features_module(
        coords, dt, physics_specs, aux_features)
    self.nodal_mapping_module = nodal_mapping_module
    self.tendency_transform_fn = tendency_transform_module(
        coords, dt, physics_specs, aux_features)
    if filter_module is not None:
      self.filter_fn = filter_module(
          coords, dt, physics_specs, aux_features)
    else:
      self.filter_fn = lambda _, y: y  # no filtering.

  def __call__(
      self,
      inputs: typing.PyTreeState,
      memory: Optional[typing.Pytree] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.Pytree] = None,
      forcing: Optional[Forcing] = None,
  ) -> typing.PyTreeState:
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    if memory is not None:
      memory, _ = pytree_utils.as_dict(memory)
    prediction_mask = self.prediction_mask
    if prediction_mask is None:
      prediction_mask = pytree_utils.tree_map_over_nonscalars(
          lambda _: True, inputs, scalar_fn=lambda _: False
      )
    prediction_shapes = jax.tree_util.tree_map(
        lambda x, y: x if y else None,
        coordinate_systems.get_nodal_shapes(inputs, self.coords),
        prediction_mask,
    )
    net = self.nodal_mapping_module(prediction_shapes)
    nodal_inputs = self.modal_to_nodal_features_fn(
        inputs, memory=memory, diagnostics=diagnostics, randomness=randomness,
        forcing=forcing,
    )
    nodal_tendencies = net(nodal_inputs)
    nodal_tendencies = self.tendency_transform_fn(nodal_tendencies)
    modal_tendencies = self.coords.horizontal.to_modal(nodal_tendencies)
    modal_tendencies = self.filter_fn(inputs, modal_tendencies)
    return from_dict_fn(modal_tendencies)


@gin.register
class DivCurlNeuralParameterization(hk.Module):
  """Computes modal physics tendencies via `u, v` → `δ, ζ`."""

  def __init__(
      self,
      coords: coordinate_systems.CoordinateSystem,
      dt: float,
      physics_specs: Any,
      aux_features: typing.AuxFeatures,
      modal_to_nodal_features_module: FeaturesModule,
      nodal_mapping_module: mappings.MappingModule,
      tendency_transform_module: TransformModule,
      prediction_mask: Optional[typing.Pytree] = None,
      filter_module: Optional[StepFilterModule] = None,
      name: Optional[str] = None,
  ):
    super().__init__(name=name)
    self.coords = coords
    self.prediction_mask = prediction_mask
    self.modal_to_nodal_features_fn = modal_to_nodal_features_module(
        coords, dt, physics_specs, aux_features)
    self.nodal_mapping_module = nodal_mapping_module
    self.tendency_transform_fn = tendency_transform_module(
        coords, dt, physics_specs, aux_features)
    self.get_nodal_shape_fn = (
        lambda x: coordinate_systems.get_nodal_shapes(x, coords))
    self.to_div_curl_fn = transforms.ToModalWithDivCurlTransform(
        coords, dt, physics_specs, aux_features)
    if filter_module is not None:
      self.filter_fn = filter_module(
          coords, dt, physics_specs, aux_features)
    else:
      self.filter_fn = lambda _, y: y  # no filtering.

  def __call__(
      self,
      inputs: typing.PyTreeState,
      memory: Optional[typing.Pytree] = None,
      diagnostics: Optional[typing.Pytree] = None,
      randomness: Optional[typing.Pytree] = None,
      forcing: Optional[Forcing] = None,
  ) -> typing.PyTreeState:
    inputs = self.coords.with_dycore_sharding(inputs)
    inputs, from_dict_fn = pytree_utils.as_dict(inputs)
    if memory is not None:
      memory = self.coords.with_dycore_sharding(memory)
      memory, _ = pytree_utils.as_dict(memory)
    prediction_mask = self.prediction_mask
    if prediction_mask is None:
      prediction_mask = pytree_utils.tree_map_over_nonscalars(
          lambda _: True, inputs, scalar_fn=lambda _: False
      )
    prediction_shapes = jax.tree_util.tree_map(
        lambda x, y: self.get_nodal_shape_fn(x) if y else None,
        inputs,
        prediction_mask,
    )
    prediction_shapes['u'] = prediction_shapes.pop('divergence')
    prediction_shapes['v'] = prediction_shapes.pop('vorticity')
    net = self.nodal_mapping_module(prediction_shapes)
    nodal_inputs = self.modal_to_nodal_features_fn(
        inputs, memory=memory, diagnostics=diagnostics, randomness=randomness,
        forcing=forcing,
    )
    nodal_inputs = self.coords.dycore_to_physics_sharding(nodal_inputs)
    nodal_tendencies = net(nodal_inputs)
    nodal_tendencies = self.coords.physics_to_dycore_sharding(nodal_tendencies)
    nodal_tendencies = self.tendency_transform_fn(nodal_tendencies)
    modal_tendencies = self.to_div_curl_fn(nodal_tendencies)
    modal_tendencies = self.filter_fn(inputs, modal_tendencies)
    outputs = from_dict_fn(modal_tendencies)
    outputs = self.coords.with_dycore_sharding(outputs)
    return outputs
